import os
import re
import subprocess
from pathlib import Path
from typing import List, Union
from RPA.Desktop import utils
from RPA.Desktop.keywords import LibraryContext, keyword

if utils.is_windows():
    import winreg
else:
    winreg = None


def output(*args):
    """Run command and return output."""
    return subprocess.check_output(args).decode().strip()


class Application:
    """Container for launched application."""

    def __init__(self, name: str, args: Union[List[str], str], shell: bool = False):
        self._name = name
        self._args = args
        self._shell = shell
        self._proc = None

    def __str__(self):
        return 'Application("{name}", pid:{pid})'.format(name=self._name, pid=self.pid)

    @property
    def is_running(self):
        if not self._proc:
            return False

        return self._proc.poll() is None

    @property
    def pid(self):
        return self._proc.pid if self._proc else None

    def start(self):
        if self._proc:
            raise RuntimeError("Application already started")

        self._proc = subprocess.Popen(  # pylint: disable=consider-using-with
            self._args, shell=self._shell
        )

    def stop(self):
        if self._proc:
            self._proc.terminate()

    def wait(self, timeout=30):
        if not self._proc:
            raise RuntimeError("Application not started")

        self._proc.communicate(timeout=int(timeout))


class ApplicationKeywords(LibraryContext):
    """Keywords for starting and stopping applications."""

    def __init__(self, ctx):
        super().__init__(ctx)
        self._apps = []

    def _create_app(
        self, name: str, args: Union[List[str], str], shell: bool = False
    ) -> Application:
        cmd = " ".join(args) if not isinstance(args, str) else args
        self.logger.info("Starting application: %s", cmd)

        app = Application(name, args, shell)
        app.start()

        self._apps.append(app)
        return app

    @keyword
    def open_application(self, name_or_path: str, *args) -> Application:
        """Start a given application by name (if in PATH),
        or by path to executable.

        :param name_or_path: Name or path of application
        :param args:         Command line arguments for application
        :returns:            Application instance

        Example:

        .. code-block:: robotframework

            Open application    notepad.exe
            Open application    c:\\path\\to\\program.exe    --example-argument
        """
        name = Path(name_or_path).name
        return self._create_app(name, [name_or_path] + list(args))

    @keyword
    def open_file(self, path: str) -> Application:
        """Open a file with the default application.

        :param path: Path to file

        Example:

        .. code-block:: robotframework

            Open file    orders.xlsx
        """
        path: Path = Path(path)

        if not path.exists():
            raise FileNotFoundError(f"File does not exist: {path}")

        # TODO: Move implementations to platform-specific adapters
        if utils.is_windows():
            return self._open_default_windows(path)
        elif utils.is_linux():
            return self._open_default_linux(path)
        elif utils.is_macos():
            return self._open_default_macos(path)
        else:
            raise NotImplementedError("Not supported for current system")

    def _open_default_windows(self, path: Path):
        """Open given file with the default Windows application."""
        path = path.resolve()

        try:
            key_root = winreg.QueryValue(winreg.HKEY_CLASSES_ROOT, path.suffix)
            key_path = rf"{key_root}\shell\open\command"
            with winreg.OpenKey(winreg.HKEY_CLASSES_ROOT, key_path) as key:
                template = winreg.QueryValueEx(key, "")[0]
        except FileNotFoundError as err:
            raise ValueError(
                f"No default application associated for '{path.name}'"
            ) from err

        # Substitute variables in command, might need to implement
        # handling of more variables (or a more generic solution)
        cmd = os.path.expandvars(template)
        cmd = cmd.replace("%1", str(path))
        cmd = cmd.replace("%L", str(path))

        return self._create_app(path.name, cmd)

    def _open_default_linux(self, path: Path):
        """Open given file with the default Linux application."""
        mimetype = output("xdg-mime", "query", "filetype", path)
        applications = output("xdg-mime", "query", "default", mimetype)
        default = output("xdg-mime", "query", "default", "text/plain")

        def executable(app):
            with open(Path("/usr/share/applications/") / app, encoding="utf-8") as fd:
                matches = re.search(r"\bExec=(\S+)", fd.read(), re.MULTILINE)
                if not matches:
                    raise FileNotFoundError(f"No executable for application '{app}'")
                return matches.group(1)

        for app in applications.split(";") + default.split(";"):
            try:
                exe = executable(app)
                return self._create_app(path.name, [exe, str(path)])
            except FileNotFoundError as err:
                self.logger.info("Launching default failed: %s", err)

        raise RuntimeError(f"No default application associated for '{path.name}'")

    def _open_default_macos(self, path: Path):
        """Open given file with the default MacOS application."""
        # TODO: Find out way to get actual default application,
        #       because open does not expose the actual PID
        return self._create_app(path.name, ["open", "-W", str(path)])

    @keyword
    def close_application(self, app: Application) -> None:
        """Close given application. Needs to be started
        with this library.

        :param app: App instance

        Example:

        .. code-block:: robotframework

            # Open a file with a default application
            ${word}=    Open file    template.docx
            # Do something with Word
            Close application    ${word}

            # Open a an application with a specific executable
            ${ie}=    Open application    iexplore.exe
            # Do something with IE
            Close application    ${ie}
        """
        if app.is_running:
            app.stop()

    @keyword
    def close_all_applications(self) -> None:
        """Close all opened applications.

        Example:

        .. code-block:: robotframework

            Open file    order1.docx
            Open file    order2.docx
            Open file    order3.docx
            # Do something with Word
            Close all applications
        """
        for app in self._apps:
            if app.is_running:
                app.stop()



import functools

try:
    from Browser import Browser
except (ModuleNotFoundError, ImportError) as exc:
    raise ModuleNotFoundError(
        "Please install the `robotframework-browser` package by following these "
        "instructions: https://rpaframework.org/libraries/browser_playwright/"
        "index.html#install-instructions"
    ) from exc
from Browser.keywords import PlaywrightState

from RPA.Browser.common import auto_headless


class RobocorpPlaywrightState(PlaywrightState):
    """Automatic headless detection is supported when opening a new browser."""

    __doc__ = f"{PlaywrightState.__doc__}\n{__doc__}"

    open_browser = auto_headless(PlaywrightState.open_browser)
    new_browser = auto_headless(PlaywrightState.new_browser)
    new_persistent_context = auto_headless(PlaywrightState.new_persistent_context)


# pylint: disable=missing-class-docstring
class Playwright(Browser):
    """Automatic headless detection is supported when opening a new browser."""

    __doc__ = f"{Browser.__doc__}\n{__doc__}"

    @functools.wraps(Browser.__init__)
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._playwright_state = RobocorpPlaywrightState(self)
        library_components = [self._playwright_state]
        self.add_library_components(library_components)



import contextlib
import datetime
from io import BytesIO
from pathlib import Path

import pytest
from RPA.Excel.Files import Files, XlsxWorkbook, XlsWorkbook, ensure_unique
from RPA.Tables import Table

from . import RESOURCES_DIR, RESULTS_DIR


EXCELS_DIR = RESOURCES_DIR / "excels"


@contextlib.contextmanager
def _library(excel_file):
    lib = Files()
    excel_path = EXCELS_DIR / excel_file
    lib.open_workbook(excel_path)
    yield lib
    lib.close_workbook()


@pytest.fixture(params=["example.xlsx", "example.xls"])
def library(request):
    with _library(request.param) as lib:
        yield lib


@pytest.fixture(params=["one-row.xlsx", "one-row.xls", "empty.xlsx", "empty.xls"])
def library_empty(request):
    with _library(request.param) as lib:
        yield lib


@pytest.mark.parametrize(
    "fmt, instance", [("xlsx", XlsxWorkbook), ("xls", XlsWorkbook)]
)
def test_create_workbook(fmt, instance):
    library = Files()
    library.create_workbook(fmt=fmt)
    assert isinstance(library.workbook, instance)
    assert library.workbook._book is not None
    assert library.workbook.extension is None


@pytest.mark.parametrize("fmt", ["xlsx", "xls"])
def test_create_after_close(fmt):
    library = Files()
    library.create_workbook(fmt=fmt)
    library.close_workbook()
    library.create_workbook(fmt=fmt)


@pytest.mark.parametrize("fmt", ["xlsx", "xls"])
def test_create_without_close(fmt):
    library = Files()
    library.create_workbook(fmt=fmt)
    library.create_workbook(fmt=fmt)


@pytest.mark.parametrize("filename", ["not-a-file.xlsx", "not-a-file.xls"])
def test_open_missing(filename):
    with pytest.raises(FileNotFoundError):
        lib = Files()
        lib.open_workbook(filename)


def test_wrong_extension_fallback_xlsx():
    # openpyxl does not support xls (actual format) but xlrd will succeed
    library = Files()
    path = str(EXCELS_DIR / "wrong_extension.xlsx")
    library.open_workbook(path)
    assert library.workbook is not None


def test_wrong_extension_fallback_xls():
    # openpyxl will refuse to read wrong extension and xlrd does not support xlsx
    library = Files()
    path = str(EXCELS_DIR / "wrong_extension.xls")
    with pytest.raises(ValueError, match=".*wrong_extension.xls.*path.*extension.*"):
        library.open_workbook(path)
    assert library.workbook is None


def test_extension_property(library):
    assert library.workbook.extension == Path(library.workbook.path).suffix


def test_save_workbook(library):
    fd = BytesIO()
    library.save_workbook(fd)
    # Exact size is unclear because some formatting, meta-data, etc.
    # might get stripped out
    content = fd.getvalue()
    assert len(content) > 1024


def test_list_worksheets(library):
    sheets = library.list_worksheets()
    assert sheets == ["First", "Second"]


def test_get_active_worksheet(library):
    active = library.get_active_worksheet()
    assert active == "Second"


def test_set_active_worksheet(library):
    library.set_active_worksheet("First")
    assert library.workbook.active == "First"


def test_set_active_worksheet_unknown(library):
    with pytest.raises(ValueError):
        library.set_active_worksheet("Third")
    assert library.workbook.active != "Third"


def test_create_worksheet(library):
    library.create_worksheet("New")
    assert library.list_worksheets() == ["First", "Second", "New"]


def test_create_worksheet_duplicate(library):
    library.create_worksheet("New")
    assert library.list_worksheets() == ["First", "Second", "New"]

    with pytest.raises(ValueError):
        library.create_worksheet("New")
    assert library.list_worksheets() == ["First", "Second", "New"]


def test_create_worksheet_content(library):
    table = Table({"one": [1, 2, 3], "two": ["a", "b", "c"]})

    library.create_worksheet("New", table)
    data = library.read_worksheet_as_table()
    assert len(data) == 3
    assert data.get_column("A", as_list=True) == [1, 2, 3]

    library.create_worksheet("New2", table, header=True)
    data = library.read_worksheet_as_table(header=True)
    assert len(data) == 3
    assert data.get_column("one", as_list=True) == [1, 2, 3]


def test_read_worksheet_default(library):
    data = library.read_worksheet()
    assert len(data) == 10
    assert data[5]["A"] == 5
    assert data[5]["C"] == 2468


def test_read_worksheet_by_index(library):
    data = library.read_worksheet(0)
    assert len(data) == 10
    assert data[2]["B"] == "Mara"
    assert data[2]["F"] == 25


def test_read_worksheet_by_name(library):
    data = library.read_worksheet("First")
    assert len(data) == 10
    assert data[2]["B"] == "Mara"
    assert data[2]["F"] == 25


def test_read_worksheet_header(library):
    data = library.read_worksheet("Second", header=True)
    assert len(data) == 9
    assert data[5]["Index"] == 6
    assert data[5]["Id"] == 2554


@pytest.mark.parametrize(
    "header, content",
    [
        (False, [{"A": "Single"}]),
        (True, []),
    ],
)
def test_read_worksheet_header_empty(library_empty, header, content):
    data = library_empty.read_worksheet("Sheet", header=header)
    excel_name = library_empty.workbook.path.stem
    if "empty" in excel_name:
        content = []  # there's no content at all, no matter the header switch
    assert data == content


def test_read_worksheet_timestamp(library):
    data = library.read_worksheet(name="Second", header=True)
    assert data[5]["Date"] == datetime.datetime(2015, 5, 21)


def test_read_worksheet_as_table(library):
    table = library.read_worksheet_as_table(name="First")
    assert isinstance(table, Table)
    assert len(table) == 10
    assert table[2, 1] == "Mara"
    assert table[2, 2] == "Hashimoto"


def test_read_worksheet_as_table_start_offset(library):
    table = library.read_worksheet_as_table(name="First", start=3)
    assert len(table) == 8
    assert table[0, 1] == "Mara"
    assert table[0, 2] == "Hashimoto"


def test_read_worksheet_as_table_start_offset_and_header(library):
    table = library.read_worksheet_as_table(name="First", start=2, header=True)
    assert len(table) == 8
    assert table.columns == ["1", "Dulce", "Abril", "Female", "United States", "32"]
    assert table[0, 2] == "Hashimoto"


def test_read_worksheet_empty(library):
    library.create_worksheet("Empty")

    data = library.read_worksheet(header=False)
    assert data == []

    data_header = library.read_worksheet(header=True)
    assert data_header == []

    table = library.read_worksheet_as_table(header=False)
    assert table.dimensions == (0, 0)

    table_header = library.read_worksheet_as_table(header=True)
    assert table_header.dimensions == (0, 0)


def test_append_to_worksheet_headers(library):
    table = Table(
        [
            {"Index": 98, "Date": "today", "Id": "some_value"},
            {"Index": 99, "Date": "tomorrow", "Id": "another_value"},
        ]
    )
    library.append_rows_to_worksheet(table, header=True)

    result = library.read_worksheet_as_table(header=True)
    assert len(result) == 11
    assert result[-1] == [99, "tomorrow", "another_value"]


@pytest.mark.parametrize("fmt", ("xlsx", "xls"))
def test_append_to_worksheet_empty(fmt):
    table = Table(
        [
            {"Index": 98, "Date": "today", "Id": "some_value"},
            {"Index": 99, "Date": "tomorrow", "Id": "another_value"},
        ]
    )
    library = Files()
    library.create_workbook(fmt=fmt)
    library.append_rows_to_worksheet(table)

    result = library.read_worksheet_as_table()
    assert len(result) == 2
    assert result[0] == [98, "today", "some_value"]


@pytest.mark.parametrize("fmt", ("xlsx", "xls"))
def test_append_to_worksheet_empty_with_headers(fmt):
    table = Table(
        [
            {"Index": 98, "Date": "today", "Id": "some_value"},
            {"Index": 99, "Date": "tomorrow", "Id": "another_value"},
        ]
    )
    library = Files()
    library.create_workbook(fmt=fmt)
    library.append_rows_to_worksheet(table, header=True)

    result = library.read_worksheet_as_table()
    assert len(result) == 3
    assert result[0] == ["Index", "Date", "Id"]


def test_remove_worksheet(library):
    library.set_active_worksheet("Second")

    library.remove_worksheet("Second")
    assert library.list_worksheets() == ["First"]
    assert library.get_active_worksheet() == "First"

    with pytest.raises(ValueError):
        library.remove_worksheet("First")


def test_rename_worksheet(library):
    library.rename_worksheet("Second", "Toinen")
    assert library.list_worksheets() == ["First", "Toinen"]


def test_ensure_unique():
    result = ensure_unique(["Banana", "Apple", "Lemon", "Apple", "Apple", "Banana"])
    assert result == ["Banana", "Apple", "Lemon", "Apple_2", "Apple_3", "Banana_2"]


def test_ensure_unique_nested():
    result = ensure_unique(["Banana", "Apple", "Lemon", "Apple", "Apple_2", "Banana"])
    assert result == ["Banana", "Apple", "Lemon", "Apple_2", "Apple_2_2", "Banana_2"]


def test_find_empty_row(library):
    row = library.find_empty_row()
    assert row == 11


def test_get_worksheet_value(library):
    assert library.get_worksheet_value(5, "A") == 4
    assert library.get_worksheet_value(5, "C") == 3549
    assert library.get_worksheet_value(3, 3) == 1582
    assert library.get_worksheet_value(9, "E", "First") == "United States"


def test_set_worksheet_value(library):
    library.set_worksheet_value(11, "A", "First")
    library.set_worksheet_value(11, 2, "Second")
    library.set_worksheet_value(11, "3", "Third")

    data = library.read_worksheet()

    row = data[-1]
    assert row["A"] == "First"
    assert row["B"] == "Second"
    assert row["C"] == "Third"


def test_get_cell_value(library):
    assert library.get_cell_value(5, "A") == 4
    assert library.get_cell_value(5, "C") == 3549
    assert library.get_cell_value(3, 3) == 1582
    assert library.get_cell_value(9, "E", "First") == "United States"


def test_set_cell_value(library):
    library.set_cell_value(11, "A", "First")
    library.set_cell_value(11, 2, "Second")
    library.set_cell_value(11, "3", "Third", fmt="00.0")

    data = library.read_worksheet()

    row = data[-1]
    assert row["A"] == "First"
    assert row["B"] == "Second"
    assert row["C"] == "Third"


def test_cell_format(library):
    fmts = [
        "general",
        "0",
        "0.00",
        "#,##0",
        "#,##0.00",
        '"$"#,##0_);("$"#,##',
        '"$"#,##0_);[Red]("$"#,##',
        '"$"#,##0.00_);("$"#,##',
        '"$"#,##0.00_);[Red]("$"#,##',
        "0%",
        "0.00%",
        "0.00E+00",
        "# ?/?",
        "# ??/??",
        "M/D/YY",
        "D-MMM-YY",
        "D-MMM",
        "MMM-YY",
        "h:mm AM/PM",
        "h:mm:ss AM/PM",
        "h:mm",
        "h:mm:ss",
        "M/D/YY h:mm",
        "_(#,##0_);(#,##0)",
        "_(#,##0_);[Red](#,##0)",
        "_(#,##0.00_);(#,##0.00)",
        "_(#,##0.00_);[Red](#,##0.00)",
        '_("$"* #,##0_);_("$"* (#,##0);_("$"* "-"_);_(@_)',
        '_(* #,##0_);_(* (#,##0);_(* "-"_);_(@_)',
        '_("$"* #,##0.00_);_("$"* (#,##0.00);_("$"* "-"??_);_(@_)',
        '_(* #,##0.00_);_(* (#,##0.00);_(* "-"??_);_(@_)',
        "mm:ss",
        "[h]:mm:ss",
        "mm:ss.0",
        "##0.0E+0",
        "@",
        "BOOLEAN",
    ]

    value = -1278.9078
    library.create_worksheet("Formats")

    for idx, fmt in enumerate(fmts, 1):
        library.set_cell_value(idx, "A", fmt)
        library.set_cell_value(idx, "B", value)
        library.set_cell_format(idx, "B", fmt)


def test_insert_image_to_worksheet(library):
    path = str(RESOURCES_DIR / "faces.jpeg")
    library.insert_image_to_worksheet(10, "B", path, scale=4)
    library.save_workbook(BytesIO())


@pytest.mark.parametrize("fmt", ("xlsx", "xls"))
def test_create_workbook_default_sheet(fmt):
    library = Files()

    library.create_workbook(fmt=fmt)
    assert library.list_worksheets() == ["Sheet"]

    library.create_worksheet("Test")
    assert library.list_worksheets() == ["Sheet", "Test"]


@pytest.mark.parametrize(
    "excel_file, data_only",
    [
        ("formulas.xlsx", False),
        ("formulas.xls", False),
        ("formulas.xlsx", True),
        ("formulas.xls", True),
    ],
)
def test_read_worksheet_with_formulas(excel_file, data_only):
    library = Files()
    excel_path = EXCELS_DIR / excel_file
    library.open_workbook(excel_path, data_only=data_only)
    assert library.get_worksheet_value(2, "A") == 1
    assert library.get_worksheet_value(2, "B") == 3
    if library.workbook.path.suffix == ".xlsx":
        assert library.get_worksheet_value(2, "C") == 4 if data_only else "=A2+B2"
    else:
        assert library.get_worksheet_value(2, "C") == 4
    library.close_workbook()


@pytest.mark.parametrize("name", ["spaces.xls", "spaces.xlsx"])
def test_invalid_whitespace_fix(name):
    library = Files()
    if name.endswith("xlsx"):
        get_user = lambda book: book.properties.lastModifiedBy
        expected_user = "cmin  "
    else:
        get_user = lambda book: book.user_name
        expected_user = "cmin"

    library.open_workbook(EXCELS_DIR / name)
    assert get_user(library.workbook.book) == expected_user

    library.save_workbook(RESULTS_DIR / name)
    # Leading/trailing whitespace is stripped on save, thus not creating any unwanted
    #  `xml:space="preserve"` tag child under workbook properties. (which breaks
    #  validation with Microsoft)
    assert get_user(library.workbook.book) == "cmin"


@pytest.mark.parametrize("fmt", ["xlsx", "xls"])
def test_create_with_sheet_name(fmt):
    library = Files()
    path = RESULTS_DIR / f"custom-sheet-name.{fmt}"
    name = "CustomName"
    library.create_workbook(path, fmt=fmt, sheet_name=name)
    assert library.get_active_worksheet() == name



# pylint: disable=cyclic-import
from robot.api.deco import keyword
from .context import (
    LibraryContext,
    ElementNotFound,
    MultipleElementsFound,
    TimeoutException,
    HAS_RECOGNITION,
)

from .application import ApplicationKeywords
from .clipboard import ClipboardKeywords
from .finder import FinderKeywords
from .keyboard import KeyboardKeywords
from .mouse import MouseKeywords
from .screen import ScreenKeywords
from .text import TextKeywords



from typing import Dict, List

from robocorp import storage
from robot.api.deco import keyword, library


@library
class Storage:
    """Control Room `Asset Storage` library operating with the cloud built-in key-value
    store.

    Library requires at the minimum `rpaframework` version **24.0.0**.

    **Usage**

    .. code-block:: robotframework

        *** Tasks ***
        Manage Assets
            @{assets} =    List Assets
            Log List    ${assets}

            Set Text Asset    my-asset    My string asset value
            ${value} =      Get Text Asset       my-asset
            Log     Asset value: ${value}

            Delete Asset    my-asset

    .. code-block:: python

        import logging
        from RPA.Robocorp.Storage import Storage

        storage = Storage()

        def manage_assets():
            assets = storage.list_assets()
            logging.info(assets)

            storage.set_text_asset("my-asset", "My string asset value")
            value = storage.get_text_asset("my-asset")
            logging.info("Asset value: %s", value)

            storage.delete_asset("my-asset")

    **Caveats**

    Currently, there's no local file adapter support, therefore you need to be linked
    to Control Room and connected to a Workspace in VSCode before being able to develop
    locally robots using this functionality.

    While the content type can be controlled (during bytes and file setting), it is
    currently disabled in this version of the library for simplicity reasons.
    """

    ROBOT_LIBRARY_SCOPE = "GLOBAL"
    ROBOT_LIBRARY_DOC_FORMAT = "REST"

    @keyword
    def list_assets(self) -> List[str]:
        """List all the existing assets.

        :returns: A list of available assets' names

        **Example: Robot Framework**

        .. code-block:: robotframework

            *** Tasks ***
            Print All Assets
                @{assets} =    List Assets
                Log List    ${assets}

        **Example: Python**

        .. code-block:: python

            def print_all_assets():
                print(storage.list_assets())
        """
        return storage.list_assets()

    @keyword
    def set_bytes_asset(self, name: str, value: bytes, wait: bool = True):
        """Creates or updates an asset named `name` with the provided bytes `value`.

        :param name: Name of the existing or new asset to create (if missing)
        :param value: The new bytes value to set within the asset
        :param wait: Wait for the value to be set successfully

        **Example: Robot Framework**

        .. code-block:: robotframework

            *** Tasks ***
            Set An Asset
                ${random_bytes} =      Evaluate    os.urandom(10)      modules=os
                Set Bytes Asset    my-bytes-asset    ${random_bytes}

        **Example: Python**

        .. code-block:: python

            import os

            def set_an_asset():
                random_bytes = os.urandom(10)
                storage.set_bytes_asset("my-bytes-asset", random_bytes)
        """
        storage.set_bytes(name, data=value, wait=wait)

    @keyword
    def get_bytes_asset(self, name: str) -> bytes:
        """Get the asset's bytes value by providing its `name`.

        :param name: Name of the asset
        :raises AssetNotFound: Asset with the given name does not exist
        :returns: The current value of this asset as bytes

        **Example: Robot Framework**

        .. code-block:: robotframework

            *** Tasks ***
            Retrieve An Asset
                ${value} =      Get Bytes Asset       my-bytes-asset
                Log     Asset bytes value: ${value}

        **Example: Python**

        .. code-block:: python

            def retrieve_an_asset():
                value = storage.get_bytes_asset("my-bytes-asset")
                print(b"Asset bytes value:", value)
        """
        return storage.get_bytes(name)

    @keyword
    def set_text_asset(self, name: str, value: str, wait: bool = True):
        """Creates or updates an asset named `name` with the provided text `value`.

        :param name: Name of the existing or new asset to create (if missing)
        :param value: The new text value to set within the asset
        :param wait: Wait for the value to be set successfully

        **Example: Robot Framework**

        .. code-block:: robotframework

            *** Tasks ***
            Set An Asset
                Set Text Asset    my-text-asset    My asset value as text

        **Example: Python**

        .. code-block:: python

            def set_an_asset():
                storage.set_text_asset("my-text-asset", "My asset value as text")
        """
        storage.set_text(name, text=value, wait=wait)

    @keyword
    def get_text_asset(self, name: str) -> str:
        """Get the asset's text value by providing its `name`.

        :param name: Name of the asset
        :raises AssetNotFound: Asset with the given name does not exist
        :returns: The current value of this asset as text

        **Example: Robot Framework**

        .. code-block:: robotframework

            *** Tasks ***
            Retrieve An Asset
                ${value} =      Get Text Asset       my-text-asset
                Log     Asset text value: ${value}

        **Example: Python**

        .. code-block:: python

            def retrieve_an_asset():
                value = storage.get_text_asset("my-text-asset")
                print("Asset text value:", value)
        """
        return storage.get_text(name)

    @keyword("Set JSON Asset")
    def set_json_asset(self, name: str, value: Dict, wait: bool = True):
        """Creates or updates an asset named `name` with the provided dictionary
        `value`.

        :param name: Name of the existing or new asset to create (if missing)
        :param value: The new dictionary value to set within the asset
        :param wait: Wait for the value to be set successfully

        **Example: Robot Framework**

        .. code-block:: robotframework

            *** Tasks ***
            Set An Asset
                &{entries} =    Create Dictionary   dogs    ${10}
                Set JSON Asset    my-json-asset    ${entries}

        **Example: Python**

        .. code-block:: python

            def set_an_asset():
                entries = {"dogs": 10}
                storage.set_json_asset("my-json-asset", entries)
        """
        storage.set_json(name, value=value, wait=wait)

    @keyword("Get JSON Asset")
    def get_json_asset(self, name: str) -> Dict:
        """Get the asset's dictionary value by providing its `name`.

        :param name: Name of the asset
        :raises AssetNotFound: Asset with the given name does not exist
        :returns: The current value of this asset as a dictionary

        **Example: Robot Framework**

        .. code-block:: robotframework

            *** Tasks ***
            Retrieve An Asset
                &{value} =      Get JSON Asset       my-json-asset
                Log     Asset dictionary value: ${value}

        **Example: Python**

        .. code-block:: python

            def retrieve_an_asset():
                value = storage.get_json_asset("my-json-asset")
                print("Asset dictionary value:", value)
        """
        return storage.get_json(name)

    @keyword
    def set_file_asset(self, name: str, path: str, wait: bool = True):
        """Creates or updates an asset named `name` with the content of the given
        `path` file.

        :param name: Name of the existing or new asset to create (if missing)
        :param path: The file path whose content to be set within the asset
        :param wait: Wait for the value to be set successfully

        **Example: Robot Framework**

        .. code-block:: robotframework

            *** Tasks ***
            Set An Asset
                Set File Asset    my-file-asset      report.pdf

        **Example: Python**

        .. code-block:: python

            def set_an_asset():
                storage.set_file_asset("my-file-asset", "report.pdf")
        """
        storage.set_file(name, path=path, wait=wait)

    @keyword
    def get_file_asset(self, name: str, path: str, overwrite: bool = False) -> str:
        """Get the asset's content saved to disk by providing its `name`.

        :param name: Name of the asset
        :param path: Destination path to the downloaded file
        :param overwrite: Replace destination file if it already exists (default False)
        :raises AssetNotFound: Asset with the given name does not exist
        :returns: A local path pointing to the retrieved file

        **Example: Robot Framework**

        .. code-block:: robotframework

            *** Tasks ***
            Retrieve An Asset
                ${path} =      Get File Asset       my-file-asset    report.pdf
                Log     Asset file path: ${path}

        **Example: Python**

        .. code-block:: python

            def retrieve_an_asset():
                path = storage.get_file_asset("my-file-asset", "report.pdf")
                print("Asset file path:", path)
        """
        return str(storage.get_file(name, path=path, exist_ok=overwrite))

    @keyword
    def delete_asset(self, name: str):
        """Delete an asset by providing its `name`.

        This operation cannot be undone.

        :param name: Name of the asset to delete
        :raises AssetNotFound: Asset with the given name does not exist

        **Example: Robot Framework**

        .. code-block:: robotframework

            *** Tasks ***
            Remove An Asset
                Delete Asset    my-asset

        **Example: Python**

        .. code-block:: python

            def remove_an_asset():
                storage.delete_asset("my-asset")
        """
        storage.delete_asset(name)



import functools
import inspect
import logging
import os
import platform
from typing import Union


AUTO = "AUTO"
logger = logging.getLogger(__name__)


def get_headless_state(headless: Union[bool, str] = AUTO) -> bool:
    """Parse argument for headless mode."""
    if str(headless).strip().lower() == AUTO.lower():
        # If in Linux and with no valid display, we can assume we are in a
        #  container which doesn't support UI.
        headless = platform.system() == "Linux" and not (
            os.environ.get("DISPLAY") or os.environ.get("WAYLAND_DISPLAY")
        )
        if headless:
            logger.info("Auto-detected headless environment!")
    headless = headless or int(os.getenv("RPA_HEADLESS_MODE", "0"))
    return bool(headless)


def auto_headless(super_func):
    """Automatically handles the headless switch in a keyword."""
    signature = inspect.signature(super_func)
    default_headless = signature.parameters["headless"].default

    @functools.wraps(super_func)
    def wrapper(*args, **kwargs):
        headless = kwargs.pop("headless", default_headless)
        headless = get_headless_state(headless)
        return super_func(*args, headless=headless, **kwargs)

    return wrapper



from collections import OrderedDict
import datetime
import getpass
import logging
import os
import platform
import signal
import socket
from typing import Any

from RPA.core.decorators import operating_system_required

if platform.system() == "Windows":
    import psutil
    from psutil._common import bytes2human
else:
    psutil = object
    bytes2human = object


class OperatingSystem:
    """`OperatingSystem` is a cross-platform library for managing
    computer properties and actions.

    **Examples**

    **Robot Framework**

    .. code-block:: robotframework

        *** Settings ***
        Library    RPA.Desktop.OperatingSystem

        *** Tasks ***
        Get computer information
            ${boot_time}=   Get Boot Time  as_datetime=${TRUE}
            ${machine}=     Get Machine Name
            ${username}=    Get Username
            &{memory}=      Get Memory Stats
            Log Many        ${memory}

    **Python**

    .. code-block:: python

        from RPA.Desktop.OperatingSystem import OperatingSystem

        def get_computer_information():
            ops = OperatingSystem()
            print(f"Boot time    : { ops.get_boot_time(as_datetime=True) }"
                  f"Machine name : { ops.get_machine_name() }"
                  f"Username     : { ops.get_username() }"
                  f"Memory       : { ops.get_memory_stats() }")

        if __name__ == "__main__":
            get_computer_information()

    """

    ROBOT_LIBRARY_SCOPE = "GLOBAL"
    ROBOT_LIBRARY_DOC_FORMAT = "REST"

    def __init__(self):
        self.logger = logging.getLogger(__name__)

    @operating_system_required("Windows")
    def get_boot_time(
        self, as_datetime: bool = False, datetime_format: str = "%Y-%m-%d %H:%M:%S"
    ) -> str:
        """Get computer boot time in seconds from Epoch or in datetime string.

        :param as_datetime: if True returns datetime string, otherwise seconds,
            defaults to False
        :param datetime_format: datetime string format, defaults to "%Y-%m-%d %H:%M:%S"
        :return: seconds from Epoch or datetime string

        Example:

        .. code-block:: robotframework

            ${boottime}  Get Boot Time
            ${boottime}  Get Boot Time   as_datetime=True
            ${boottime}  Get Boot Time   as_datetime=True  datetime_format=%d.%m.%Y

        """
        btime = self.boot_time_in_seconds_from_epoch()
        if as_datetime:
            return datetime.datetime.fromtimestamp(btime).strftime(datetime_format)
        return btime

    @operating_system_required("Windows")
    def boot_time_in_seconds_from_epoch(self) -> str:
        """Get machine boot time

        :return: boot time in seconds from Epoch

        Example:

        .. code-block:: robotframework

            ${epoch}  Boot Time In Seconds From Epoch

        """
        return psutil.boot_time()

    def get_machine_name(self) -> str:
        """Get machine name

        :return: machine name as string

        Example:

        .. code-block:: robotframework

            ${machine}  Get Machine Name

        """
        return socket.gethostname()

    def get_username(self) -> str:
        """Get username of logged in user

        :return: username as string

        Example:

        .. code-block:: robotframework

            ${user}  Get Username

        """
        return getpass.getuser()

    @operating_system_required("Darwin", "Linux")
    def put_system_to_sleep(self) -> None:
        """Puts system to sleep mode

        Example:

        .. code-block:: robotframework

            Put System To Sleep

        """
        if platform.system() == "Darwin":
            os.system("pmset sleepnow")
        if platform.system() == "Linux":
            os.system("systemctl suspend")

    @operating_system_required("Windows")
    def process_exists(self, process_name: str, strict: bool = True) -> Any:
        """Check if process exists by its name

        :param process_name: search for this process
        :param strict: defines how match is made, default `True`
         which means that process name needs to be exact match
         and `False` does inclusive matching
        :return: process instance or False

        Example:

        .. code-block:: robotframework

            ${process}  Process Exists  calc
            ${process}  Process Exists  calc  strict=False

        """
        for p in psutil.process_iter():
            p_name = p.name()
            if strict and process_name.lower() == p_name.lower():
                return p
            elif not strict and process_name.lower() in p_name.lower():
                return p
        return False

    @operating_system_required("Windows")
    def process_id_exists(self, pid: int) -> Any:
        """Check if process exists by its id

        :param pid: process identifier
        :return: process instance or False

        Example:

        .. code-block:: robotframework

            ${process}  Process ID Exists  4567
            Run Keyword If   ${process}  Log  Process exists

        """
        for p in psutil.process_iter():
            if p.pid == pid:
                return p
        return False

    @operating_system_required("Windows")
    def kill_process(self, process_name: str) -> bool:
        """Kill process by name

        :param process_name: name of the process
        :return: True if succeeds False if not

        Example:

        .. code-block:: robotframework

            ${process}  Process Exists  calc  strict=False
            ${status}   Kill Process    ${process.name()}

        """
        p = self.process_exists(process_name)
        if p:
            p.terminate()
            return True
        return False

    @operating_system_required("Windows")
    def kill_process_by_pid(self, pid: int) -> None:
        """Kill process by pid

        :param pid: process identifier

        Example:

        .. code-block:: robotframework

            ${process}  Process Exists  calc  strict=False
            ${status}   Kill Process By PID   ${process.pid}

        """
        os.kill(pid, signal.SIGTERM)

    @operating_system_required("Windows")
    def get_memory_stats(self, humanized: bool = True) -> dict:
        """Get computer memory stats and return those in bytes
        or in humanized memory format.

        :param humanized: if False returns memory information in bytes, defaults to True
        :return: memory information in dictionary format

        Example:

        .. code-block:: robotframework

            &{mem}     Get Memory Stats
            &{mem}     Get Memory Stats   humanized=False

        """
        meminfo = psutil.virtual_memory()
        memdict = meminfo._asdict()
        if humanized:
            humandict = {}
            for key, val in memdict.items():
                if key == "percent":
                    humandict[key] = val
                else:
                    humandict[key] = bytes2human(val)
            return OrderedDict(humandict)
        return memdict



"""
Dummy module importing core library `RobotLogListener`,
because imports to core package should be done only by
the library itself.
"""

# flake8: noqa
# pylint: disable=unused-import
from RPA.core.logger import RobotLogListener



from dataclasses import dataclass
from enum import Enum
import functools
from itertools import count
from random import randint
from typing import Any, Dict, List, Union

from RPA.application import (
    BaseApplication,
    catch_com_error,
    to_path,
    to_str_path,
    constants,
)
from RPA.Tables import Table


def requires_workbook(func):
    """Ensures a workbook is open."""

    @functools.wraps(func)
    def wrapper(self, *args, **kwargs):
        if self.workbook is None:
            raise ValueError("No workbook open")
        return func(self, *args, **kwargs)

    return wrapper


class SearchOrder(Enum):
    """Enumeration for search order."""

    ROWS = "ROWS"
    COLUMNS = "COLUMNS"


@dataclass
class PivotField:
    """Data class for pivot field parameters."""

    data_column: str
    operation: str
    numberformat: str


def to_pivot_operation(operation_name: str):
    result = None
    if operation_name.upper() == "SUM":
        result = constants.xlSum
    elif operation_name.upper() == "AVERAGE":
        result = constants.xlAverage
    elif operation_name.upper() == "MAX":
        result = constants.xlMax
    elif operation_name.upper() == "MIN":
        result = constants.xlMin
    elif operation_name.upper() == "COUNT":
        result = constants.xlCount
    elif operation_name.upper() == "DISTINCT COUNT":
        result = constants.xlDistinctCount
    return result


def to_look_in(look_in: str):
    result = None
    if look_in.upper() == "FORMULAS":
        result = constants.xlFormulas
    elif look_in.upper() == "VALUES":
        result = constants.xlValues
    elif look_in.upper() == "COMMENTS":
        result = constants.xlComments
    elif look_in.upper() == "COMMENTS THREADED":
        result = constants.xlCommentsThreaded
    return result


def _split_rows_into_range_blocks(rows, ranges):
    range_blocks = []
    start = 0
    for r in ranges:
        end = start + r.Columns.Count
        range_blocks.append((start, end))
        start = end
    return [[row[i:j] for row in rows] for i, j in range_blocks]


def _get_rows_from_table(table_object: Table):
    rows = []
    for index in table_object.index:
        row = []
        for column in table_object._columns:  # pylint: disable=protected-access
            row.append(table_object.get_cell(index, column))
        rows.append(row)
    return rows


class Application(BaseApplication):
    """`Excel.Application` is a library for controlling the Excel application.

    **Examples**

    **Robot Framework**

    .. code-block:: robotframework

        *** Settings ***
        Library             RPA.Excel.Application
        Task Setup          Open Application
        Task Teardown       Quit Application

        *** Tasks ***
        Manipulate Excel application
            Open Workbook           workbook.xlsx
            Set Active Worksheet    sheetname=new stuff
            Write To Cells          row=1
            ...                     column=1
            ...                     value=my data
            Save Excel

        Run Excel Macro
            Open Workbook   orders_with_macro.xlsm
            Run Macro       Sheet1.CommandButton1_Click

        Export Workbook as PDF
            Open Workbook           workbook.xlsx
            Export as PDF           workbook.pdf

    **Python**

    .. code-block:: python

        from RPA.Excel.Application import Application

        app = Application()

        app.open_application()
        app.open_workbook('workbook.xlsx')
        app.set_active_worksheet(sheetname='new stuff')
        app.write_to_cells(row=1, column=1, value='new data')
        app.save_excel()
        app.quit_application()
    """

    APP_DISPATCH = "Excel.Application"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.workbook = None
        self.worksheet = None

    @property
    def _active_document(self):
        return self.workbook

    def _deactivate_document(self):
        self.workbook = None
        self.worksheet = None

    def add_new_workbook(self) -> None:
        """Adds new workbook for Excel application"""
        with catch_com_error():
            self.workbook = self.app.Workbooks.Add()

    def open_workbook(self, filename: str) -> None:
        """Open Excel by filename

        By default sets active worksheet to sheet number 1

        :param filename: path to filename
        """
        if not self._app:
            self.open_application()

        if filename.startswith("http"):
            workbook_path = filename
        else:
            workbook_path = to_path(filename)
            if not workbook_path.is_file():
                raise FileNotFoundError(f"{str(workbook_path)!r} doesn't exist")
            workbook_path = str(workbook_path)
        self.logger.info("Opening workbook: %s", workbook_path)

        with catch_com_error():
            try:
                self.workbook = self.app.Workbooks(workbook_path)
            except Exception as exc:  # pylint: disable=broad-except
                self.logger.debug(str(exc))
                self.logger.info("Trying to open workbook by another method...")
                self.workbook = self.app.Workbooks.Open(workbook_path)

        self.set_active_worksheet(sheetnumber=1)
        self.logger.debug("Current workbook: %s", self.workbook)

    @requires_workbook
    def set_active_worksheet(
        self, sheetname: str = None, sheetnumber: int = None
    ) -> None:
        """Set active worksheet by either its sheet number or name

        :param sheetname: name of Excel sheet, defaults to None
        :param sheetnumber: index of Excel sheet, defaults to None
        """
        with catch_com_error():
            if sheetnumber:
                self.worksheet = self.workbook.Worksheets(int(sheetnumber))
            elif sheetname:
                self.worksheet = self.workbook.Worksheets(sheetname)

    def add_new_sheet(self, sheetname: str, create_workbook: bool = True) -> None:
        """Add new worksheet to workbook. Workbook is created by default if
        it does not exist.

        :param sheetname: name for sheet
        :param create_workbook: create workbook if True, defaults to True
        :raises ValueError: error is raised if workbook does not exist and
            `create_workbook` is False
        """
        if not self.workbook:
            if not create_workbook:
                raise ValueError("No workbook open")
            self.add_new_workbook()

        self.logger.info("Adding sheet: %s", sheetname)
        with catch_com_error():
            last = self.app.Worksheets(self.app.Worksheets.Count)
            self.worksheet = self.app.Worksheets.Add(After=last)
            self.worksheet.Name = sheetname

    def find_first_available_row(
        self, worksheet: Any = None, row: int = 1, column: int = 1
    ) -> Any:
        """Find first available free row

        :param worksheet: worksheet to handle, defaults to active worksheet if None
        :param row: starting row for search, defaults to 1
        :param column: starting column for search, defaults to 1
        :return: row or None
        """
        cell = self.find_first_available_cell(worksheet, row, column)
        return cell

    @requires_workbook
    def find_first_available_cell(
        self, worksheet: Any = None, row: int = 1, column: int = 1
    ) -> Any:
        """Find first available free cell

        :param worksheet: worksheet to handle, defaults to active worksheet if None
        :param row: starting row for search, defaults to 1
        :param column: starting column for search, defaults to 1
        :return: tuple (row, column) or (None, None) if not found
        """
        if worksheet:
            self.set_active_worksheet(worksheet)

        with catch_com_error():
            for current_row in count(int(row)):
                cell = self.worksheet.Cells(current_row, column)
                if cell.Value is None:
                    return current_row, column

        return None, None

    @requires_workbook
    def write_to_cells(
        self,
        worksheet: Any = None,
        row: int = None,
        column: int = None,
        value: str = None,
        number_format: str = None,
        formula: str = None,
    ) -> None:
        """Write value, number_format and/or formula into cell.

        :param worksheet: worksheet to handle, defaults to active worksheet if None
        :param row: target row, defaults to None
        :param column: target row, defaults to None
        :param value: possible value to set, defaults to None
        :param number_format: possible number format to set, defaults to None
        :param formula: possible format to set, defaults to None
        :raises ValueError: if cell is not given
        """
        if row is None or column is None:
            raise ValueError("No cell was given")

        if worksheet:
            self.set_active_worksheet(worksheet)

        with catch_com_error():
            cell = self.worksheet.Cells(int(row), int(column))

            if number_format:
                cell.NumberFormat = number_format
            if value:
                cell.Value = value
            if formula:
                cell.Formula = formula

    @requires_workbook
    def read_from_cells(
        self,
        worksheet: Any = None,
        row: int = None,
        column: int = None,
    ) -> str:
        """Read value from cell.

        :param worksheet: worksheet to handle, defaults to active worksheet if None
        :param row: target row, defaults to None
        :param column: target row, defaults to None
        :raises ValueError: if cell is not given
        """
        if row is None or column is None:
            raise ValueError("No cell was given")

        if worksheet:
            self.set_active_worksheet(worksheet)

        with catch_com_error():
            cell = self.worksheet.Cells(int(row), int(column))
            return cell.Value

    @requires_workbook
    def save_excel(self) -> None:
        """Saves Excel file"""
        with catch_com_error():
            self.workbook.Save()

    @requires_workbook
    def save_excel_as(
        self, filename: str, autofit: bool = False, file_format=None
    ) -> None:
        """Save Excel with name if workbook is open

        :param filename: where to save file
        :param autofit: autofit cell widths if True, defaults to False
        :param file_format: format of file

        **Note:** Changing the file extension for the path does not
        affect the actual format. To use an older format, use
        the ``file_format`` argument with one of the following values:

        https://docs.microsoft.com/en-us/office/vba/api/excel.xlfileformat

        Examples:

        .. code-block:: robotframework

            # Save workbook in modern format
            Save excel as    orders.xlsx

            # Save workbook in Excel 97 format (format from above URL)
            Save excel as    legacy.xls   file_format=${56}
        """
        with catch_com_error():
            if autofit:
                self.worksheet.Rows.AutoFit()
                self.worksheet.Columns.AutoFit()

            path = to_str_path(filename)
            if file_format is not None:
                self.workbook.SaveAs(path, FileFormat=file_format)
            else:
                self.workbook.SaveAs(path)

    @requires_workbook
    def run_macro(self, macro_name: str, *args: Any) -> None:
        """Run Excel macro with given name

        :param macro_name: macro to run
        :param args: arguments to pass to macro
        """
        with catch_com_error():
            self.app.Application.Run(f"'{self.workbook.Name}'!{macro_name}", *args)

    def export_as_pdf(self, pdf_filename: str, excel_filename: str = None) -> None:
        """Export Excel as PDF file

        If Excel filename is not given, the currently open workbook
        will be exported as PDF.

        :param pdf_filename: PDF filename to save
        :param excel_filename: Excel filename to open
        """
        if excel_filename:
            self.open_workbook(excel_filename)
        else:
            if not self.workbook:
                raise ValueError("No workbook open, can't export PDF")
        with catch_com_error():
            pdf_path = to_str_path(pdf_filename)
            self.workbook.ExportAsFixedFormat(0, pdf_path)

    def create_pivot_field(
        self,
        data_column: str,
        operation: str,
        numberformat: str = None,
    ) -> PivotField:
        """Create pivot field object parameters.

        *Note.* At the moment operation "DISTINCT COUNT" is not
        supported as there seems to be issues in the COM interface,
        which have not been resolved yet (regarding this library
        implementation).

        Python example:

        .. code-block:: python

            field_count = excel.create_pivot_field("price", "count", "#")
            field_avg = excel.create_pivot_field("price", "average", "#0,#0")

        Robot Framework example:

        .. code-block:: robotframework

            ${field_sum}=    Create Pivot Field    price    sum    #,#0
            ${field_max}=    Create Pivot Field    price    max    #,#0

        :param data_column: name of the data column
        :param operation: name of the possible operations
         (SUM, AVERAGE, MAX, MIN, COUNT)
        :param numberformat: Excel cell number format, by default
         number format is not set for the field
        :return: field object
        """
        return PivotField(data_column, operation, numberformat)

    def create_pivot_table(
        self,
        source_worksheet: str,
        pivot_worksheet: str,
        rows: List[str],
        fields: List[PivotField],
        sort_field: PivotField = None,
        sort_direction: str = "descending",
        data_range: Any = None,
        pivot_name: str = "PivotTable1",
        collapse_rows: bool = True,
        show_grand_total: bool = True,
    ) -> Any:
        """Create a pivot table in the specified worksheet.

        This is a initial implementation of the pivot table creation,
        which might not work in all cases. The alternative way
        of creating pivot tables is to use a macro an run it.

        Python example:

        .. code-block:: python

            rows = ["products", "expense_type"]
            field_count = excel.create_pivot_field("price", "count", "#")
            field_avg = excel.create_pivot_field("price", "average", "#0,#0")
            pivottable = excel.create_pivot_table(
                source_worksheet="data",
                pivot_worksheet="test!R5C5",
                rows=rows,
                fields=[field_count, field_avg]
            )

        Robot Framework example:

        .. code-block:: robotframework

            @{rows}=    Create List    products    expense_type
            ${field_sum}=    Create Pivot Field    price    sum    #,#0
            ${field_max}=    Create Pivot Field    price    max    #,#0
            @{fields}=   Create List   ${field_sum}    ${field_max}
            ${pivottable}=    Create Pivot Table
            ...    source_worksheet=data
            ...    pivot_worksheet=test!R5C5
            ...    rows=${rows}
            ...    fields=${fields}

        :param source_worksheet: name of the source worksheet
        :param pivot_worksheet: name of the pivot worksheet, can
         be the same as the source worksheet but then cell location
         of the pivot table needs to be given in the format "R1C1"
         (R is a column numbe and C is a row number, e.g. "R1C1" is A1)
        :param rows: columns in the `source_worksheet` which are used
         as pivot table rows
        :param fields: columns for the pivot table data fields
        :param sort_field: field to sort the pivot table by (one of the
         `fields`)
        :param sort_direction: sort direction (ascending or descending),
         default is descending
        :param data_range: source data range, if not given then
         the whole used range of `source_worksheet` will be used
        :param pivot_name: name of the pivot table, if not given
         then the name is "PivotTable1"
        :param collapse_rows: if `True` then the first row will be collapsed
        :param show_grand_total: if `True` then the grand total will be shown
         for the columns
        :return: created `PivotTable` object
        """

        self.set_active_worksheet(source_worksheet)
        if data_range:
            excel_range = (
                self.get_range(data_range)
                if isinstance(data_range, str)
                else data_range
            )
        else:
            excel_range = self.worksheet.UsedRange

        # Grab the pivot table source data
        pivot_cache = self.workbook.PivotCaches().Create(
            SourceType=constants.xlDatabase,
            SourceData=excel_range,
            Version=constants.xlPivotTableVersion15,
        )

        # Create the pivot table object
        table_destination = (
            f"{pivot_worksheet}!R1C1" if "!" not in pivot_worksheet else pivot_worksheet
        )

        pivot_table = pivot_cache.CreatePivotTable(
            TableDestination=table_destination,
            TableName=pivot_name,
            ReadData=True,
            DefaultVersion=constants.xlPivotTableVersion15,
        )

        for index, row in enumerate(rows):
            row_field = pivot_table.PivotFields(row)
            row_field.Orientation = constants.xlRowField
            row_field.Position = index + 1

        if len(rows) > 0 and collapse_rows:
            pivot_table.PivotFields(rows[0]).ShowDetail = False

        for pt_field in fields:
            self.logger.info(pt_field)
            pivot_operation = to_pivot_operation(pt_field.operation)

            # Access the field from the PivotFields collection
            field = pivot_table.PivotFields(pt_field.data_column)
            # Add the field to the values area
            data_field = pivot_table.AddDataField(field)
            data_field.Orientation = constants.xlDataField
            with catch_com_error():
                data_field.Name = f"{pt_field.data_column}_{randint(1000,9999)}"
                data_field.Function = pivot_operation
                if pt_field.numberformat:
                    data_field.NumberFormat = pt_field.numberformat

        if sort_field:

            def to_sort_direction(sort_direction: str):
                if sort_direction.upper() == "ASCENDING":
                    return constants.xlAscending
                else:
                    return constants.xlDescending

            # Access the pivot field (row or column) you want to sort
            pivot_field = pivot_table.PivotFields(rows[0])

            # Access the data field (value field) you want to sort by
            item_name = (
                f"{sort_field.operation.capitalize()} of {sort_field.data_column}"
            )
            data_field = pivot_table.DataPivotField.PivotItems(item_name)

            field_sort_direction = to_sort_direction(sort_direction)
            # Sort the pivot field in descending order based on the data field
            pivot_field.AutoSort(Order=field_sort_direction, Field=data_field.Name)

        # Visiblity True or Valse
        pivot_table.ShowValuesRow = False
        pivot_table.ColumnGrand = show_grand_total
        pivot_table.RefreshTable()
        return pivot_table

    def find(
        self,
        search_string: str,
        search_range: Any = None,
        max_results: int = None,
        search_order: SearchOrder = SearchOrder.ROWS,
        match_case: bool = False,
        search_type: str = None,
        search_after: str = None,
        exact: bool = False,
    ) -> List[Any]:
        """Keyword for finding text in the current worksheet.

        Wildcard can be used in a search string. The asterisk (*) represents
        any series of characters, and the question mark (?) represents a single
        character.

        Python example:

        .. code-block:: python

            ranges = excel.find("32.145.???.1", "IP!E1:E9999", 6)
            for r in ranges:
                print(f"ADDR = {r.Address} VALUE = {r.Value}")
                r.Value = r.Value.replace("32.145.", "192.168.")
                r.BorderAround()

        Robot Framework example:

        .. code-block:: robotframework

            ${ranges}=    Find
            ...    search_string=32.145.*
            ...    search_range=IP!A1:A9999
            ...    max_results=6
            ...    search_order=COLUMNS

            FOR    ${ranges}    IN    @{ranges}
                ${value}=    Set Variable    ${range.Value}
                Log to console    ADDR = ${range.Address} VALUE = ${value}
                ${new_value}=    Replace String    ${value}    32.145.    192.168.
                Set Object Property    ${range}    Value    ${new_value}
                Call Method    ${range}    BorderAround
            END

        :param search_string: what to search for
        :param search_range: if not given will search the current
         worksheet
        :param max_results: can be used to limit number of results
        :param search_order: by default search is executed by ROWS,
         can be changed to COLUMNS
        :param match_case: if `True` then the search is case sensitive
        :param search_type: can be FORMULAS, VALUES, COMMENTS or COMMENTS THREADED
        :param search_after: search after this cell
        :param exact: if `True` then the search is expected to be a exact match
        :return: list of `Range` objects
        """
        parameters = {
            "SearchOrder": (
                constants.xlByRows
                if search_order == SearchOrder.ROWS
                else constants.xlByColumns
            ),
            "MatchCase": match_case,
            "What": search_string,
        }
        if search_type:
            parameters["LookIn"] = to_look_in(search_type)
        if search_after:
            parameters["After"] = search_after
        results = []
        if search_range:
            search_area = (
                self._app.Range(search_range)
                if isinstance(search_range, str)
                else search_range
            )
        else:
            search_area = self.worksheet.UsedRange  # self.worksheet.Cells
        found = search_area.Find(**parameters)
        if not found:
            self.logger.info(
                f"Did not find the '{search_string}' in the current "
                f"{self.worksheet.Name}' worksheet"
            )
            return results
        addresses = set()
        while found:
            if max_results and len(results) >= max_results:
                break
            if found.Address in addresses:
                break
            addresses.add(found.Address)
            if found:
                expected = search_string if match_case else search_string.lower()
                actual = found.Value if match_case else found.Value.lower()
                if (exact and actual == expected) or not exact:
                    results.append(found)
            found = search_area.FindNext(found)
        return results

    def create_table(self, table_name: str, table_range: Any = None) -> None:
        """Create a table in the current worksheet.

        :param table_name: name for the table
        :param table_range: source table range, if not given then
         the whole used range of `source_worksheet` will be used
        """
        if table_range:
            excel_range = (
                self.get_range(table_range)
                if isinstance(table_range, str)
                else table_range
            )
        else:
            excel_range = self.worksheet.UsedRange

        self.worksheet.ListObjects.Add(
            SourceType=constants.xlSrcRange,
            Source=excel_range,
            XlListObjectHasHeaders=constants.xlYes,
        ).Name = table_name

    def list_tables(self) -> List[str]:
        """Return tables in the current worksheet.

        :return: list of table names
        """
        return [table.Name for table in self.worksheet.ListObjects]

    def get_range(self, table_range: str) -> Any:
        """Get range object for the given range address.

        These object properties and methods can be then called.

        Python example:

        .. code-block:: python

            source = excel.get_range('A1:B2')
            for r in source:
                print(f"ADDR = {r.Address} VAL = {r.Value}")
                r.BorderAround()
            source.Merge()
            # Creating a formula and copying it to another range
            excel.get_range("E4").Formula = "=SUM(C4:D4)"
            destination = excel.get_range("E5:E10")
            excel.get_range("E4").Copy(destination)

        Robot Framework example:

        .. code-block:: robotframework

            ${range}=    Get Range    data!A1:A4
            FOR    ${r}    IN    @{range}
                Log To Console    ADDR = ${r.Address} VAL = ${r.Value}
                Call Method  ${r}    BorderAround
            END
            Call Method    ${range}    Merge

        :param table_range: range to return
        :return: range object
        """
        return self.worksheet.Range(table_range)

    def get_pivot_tables(
        self, pivot_table_name: str = None, as_list: bool = True
    ) -> Dict[str, Any]:
        """Return pivot tables in the current worksheet.

        Python example:

        .. code-block:: python

            from RPA.Tables import Tables

            pivot_tables = excel.get_pivot_tables()

            for tbl_name, tbl_list in pivot_tables.items():
                print(f"TABLE NAME: {tbl_name}")
                table = Tables().create_table(data=tbl_list[1:], columns=tbl_list[0])
                print(table)

        Robot Framework example:

        .. code-block:: robotframework

            ${pivots}=    Get Pivot Tables
            FOR    ${tablename}    ${pivot}    IN    &{pivots}
                Log To Console    ${tablename}
                ${table}=    RPA.Tables.Create Table
                ...   data=${{$pivot[1:]}}
                ...   columns=${{$pivot[0]}}
                Log To Console    ${table}
            END

        :param pivot_table_name: name of the pivot table to return,
         will return by default all pivot tables
        :param as_list: if `True` then the pivot table data is returned as list
         of lists, if `False` then the data is returned as list of `Range` objects
        :return: dictionary of pivot tables (names as keys and table data as values)
        """
        pivot_tables = {}
        tables = self.worksheet.PivotTables()
        for table in tables:
            data = []
            if as_list:
                previous_row = -1
                row = []
                for r in table.TableRange1:
                    if previous_row == -1 or previous_row != r.Row:
                        if len(row) > 0:
                            data.append(row)
                        row = [r.Value]
                        previous_row = r.Row
                    else:
                        row.append(r.Value)
            else:
                data = table.TableRange1
            if pivot_table_name and table.Name != pivot_table_name:
                continue
            pivot_tables[table.Name] = data
        return pivot_tables

    def write_data_to_range(
        self,
        target_range: Any,
        values: Union[Table, List[List]],
        log_warnings: bool = True,
    ):
        """Writes data to the specified range(s) in the Excel worksheet.

        The range width should match the number of columns in the data.

        Multiple ranges can be specified by separating them with a semicolon, but
        still the total width of ranges should match the number of columns in the data.

        Python example:

        .. code-block:: python

            from RPA.Tables import Tables
            from RPA.Excel.Application import Application

            excel = Application()

            table = Tables().read_table_from_csv("input.csv", header=True)
            excel.open_workbook("result.xslx)
            excel.write_data_to_range("A2:P100", table)

        Robot Framework example:

        .. code-block:: robotframework

            ${input_table}=    Read table from CSV    input.csv    header=True
            Open Workbook      result.xlsx
            Write Data To Range    A2:L21    ${input_table}   # Single range
            Write Data To Range    C2:E21;G2:I21    ${input_table}   # Multiple ranges

        :param target_range: A1 string presentation of the range(s) to write or
         Range object.
        :param values: `Table` or list of lists to write to the range(s).
        :param log_warnings: on `False` will suppress logging warning, default
         is `True` (warnings are logged)
        """
        try:
            if isinstance(target_range, str):
                target_ranges = target_range.split(";")
                ranges = [self.worksheet.Range(range) for range in target_ranges]
            else:
                ranges = [target_range]
        except Exception as err:
            raise AttributeError(
                "Unable to form valid Excel range(s) from "
                f"incoming 'target_range': f{str(err)}"
            ) from err
        rows = _get_rows_from_table(values) if isinstance(values, Table) else values
        datas = _split_rows_into_range_blocks(rows, ranges)
        row_columns = len(rows[0])
        self.logger.info(f"Data contains {len(rows)} rows and {row_columns} columns.")
        range_columns = 0
        for range_, data in zip(ranges, datas):
            columns = range_.Columns.Count
            range_columns += columns
            self.logger.info(
                f"Range contains {range_.Rows.Count} rows and {columns} columns."
            )
            range_.Value = data
        if log_warnings and range_columns != row_columns:
            self.logger.warning(
                f"Total range column count {range_columns} "
                f"is different from data to write column count {row_columns}"
            )

    def remove_hidden_columns_and_rows(
        self, initial_range: Any, worksheet: str = None
    ) -> Any:
        """Removes hidden columns and rows from a range and returns a new range.

        :param initial_range: range of cells to remove hidden columns and rows from
        :param worksheet: set active worksheet (name) before removing hidden
         columns and rows
        :return: new range or initial range if no hidden cells found
        """
        initial_range = (
            self.get_range(initial_range)
            if isinstance(initial_range, str)
            else initial_range
        )
        if worksheet:
            self.set_active_worksheet(worksheet)
        try:
            visible_range = initial_range.SpecialCells(constants.xlCellTypeVisible)
            return visible_range
        except Exception as e:  # pylint: disable=broad-except
            self.logger.error(f"No visible cells found or an error occurred:f{str(e)}")
        return initial_range

    def unmerge_range(self, initial_range: Any) -> None:
        """Unmerges a range of cells.

        :param initial_range: range of cells to unmerge
        """
        initial_range = (
            self.get_range(initial_range)
            if isinstance(initial_range, str)
            else initial_range
        )
        initial_range.UnMerge()

    def merge_range(self, initial_range: Any) -> None:
        """Merges a range of cells.

        :param initial_range: range of cells to merge
        """
        initial_range = (
            self.get_range(initial_range)
            if isinstance(initial_range, str)
            else initial_range
        )
        initial_range.Merge()



import logging

from RPA.Browser.Selenium import Selenium as _Selenium


class Browser(_Selenium):  # pylint: disable=missing-class-docstring
    __doc__ = _Selenium.__doc__

    def __init__(self, *args, **kwargs):
        logging.warning(
            "This is a deprecated import that will "
            "be removed in favor of RPA.Browser.Selenium"
        )
        super().__init__(*args, **kwargs)



from ftplib import FTP as FTPconn
from ftplib import FTP_TLS as TLSconn
from ftplib import all_errors, error_perm
from functools import wraps
import logging
import os
import ssl
from typing import Tuple
from RPA.core.notebook import notebook_file


class AuthenticationException(Exception):
    """Raised when server authentication fails"""


class FTPException(Exception):
    """Raised on general FTP error"""


def ftpcommand(f):
    @wraps(f)
    def wrapper(*args, **kwargs):
        if args[0].instance is None:
            raise FTPException("No FTP connection")
        try:
            ret = f(*args, **kwargs)
            return ret
        except FileNotFoundError as e:
            args[0].logger.warning(str(e))
            return False
        except error_perm as e:
            args[0].logger.warning(str(e))
            return False
        except all_errors as e:
            raise FTPException from e

    return wrapper


class ImplicitFTP_TLS(TLSconn):
    """FTP_TLS subclass that automatically wraps sockets in
    SSL to support implicit FTPS."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._sock = None

    @property
    def sock(self):
        """Return the socket."""
        return self._sock

    @sock.setter
    def sock(self, value):
        """When modifying the socket, ensure that it is ssl wrapped."""
        if value is not None and not isinstance(value, ssl.SSLSocket):
            value = self.context.wrap_socket(value)
        self._sock = value


class FTP:
    """`FTP` library can be used to access an FTP server,
    and interact with files.

    The library is based on Python's built-in `ftplib`_.

    .. _ftplib: https://docs.python.org/3/library/ftplib.html

    **Examples**

    **Robot Framework**

    .. code-block:: robotframework

        *** Settings ***
        Library    RPA.FTP

        *** Variables ***
        ${HOST}       127.0.0.1
        ${PORT}       27345
        ${USER}       user
        ${PASS}       12345

        *** Tasks ***
        List files on the server directory
            Connect   ${HOST}  ${PORT}  ${USER}  ${PASS}
            @{files}  List Files
            FOR  ${file}  IN  @{files}
                Log  ${file}
            END

    **Python**

    .. code-block:: python

        from RPA.FTP import FTP

        library = FTP()
        library.connect('127.0.0.1', 27345, 'user', '12345')
        files = library.list_files()
        for f in files:
            print(f)
    """

    ROBOT_LIBRARY_SCOPE = "GLOBAL"
    ROBOT_LIBRARY_DOC_FORMAT = "REST"

    def __init__(self):
        self.instance = None
        self.logger = logging.getLogger(__name__)

    def connect(
        self,
        host: str,
        port: int = 21,
        user: str = None,
        password: str = None,
        tls: bool = False,
        transfer: str = "passive",
        keyfile: str = None,
        certfile: str = None,
        timeout: int = None,
        source_address: Tuple[str, int] = None,
    ) -> bool:
        """Connect to FTP server

        :param host: address of the server
        :param port: port of the server, defaults to 21
        :param user: login name, defaults to None
        :param password: login password, defaults to None
        :param tls: connect using TLS support, defaults to False
        :param transfer: mode of the transfer, defaults to "passive"
        :param keyfile: path to private key file
        :param certfile: path to certificate file
        :param timeout: a timeout in seconds for the connection attempt
        :param source_address: socket to bind to as its source address before connecting
        :raises AuthenticationException: on authentication error with the server
        :return: true if the connnection completes and an error code if it fails
        """
        try:
            if tls:
                self.instance = ImplicitFTP_TLS(
                    keyfile=keyfile,
                    certfile=certfile,
                    timeout=timeout,
                    source_address=source_address,
                )
            else:
                self.instance = FTPconn(timeout=timeout, source_address=source_address)
            self.instance.connect(host, port)
            if user and password:
                self.instance.login(user=user, passwd=password)
            else:
                self.instance.login()
            if tls:
                self.instance.prot_p()
            if transfer != "passive":
                self.instance.set_pasv(False)
        except error_perm as e:
            raise AuthenticationException from e
        except all_errors as e:
            raise FTPException from e

        self.logger.info("FTP connection successful")
        return True

    def quit(self) -> None:
        """Send QUIT command to the server and close connection"""
        try:
            self.instance.quit()
        except all_errors as e:
            self.logger.debug(str(e))
            self.close()
        finally:
            self.instance = None

    def close(self) -> None:
        """Close connection to the server unilaterally"""
        if self.instance:
            self.instance.close()
            self.instance = None

    @ftpcommand
    def upload(self, localfile: str, remotefile: str) -> bool:
        """Upload file to FTP server

        :param localfile: path to file to upload
        :param remotefile: name of uploaded file in the server
        :return: true or false based on success or failure
        """
        cmd = f"STOR {remotefile}"
        with open(localfile, "rb") as lfile:
            self.instance.storbinary(cmd, lfile)

    def download(self, remotefile: str, localfile: str = None) -> bool:
        """Download file from FTP server

        :param remotefile: path to remote file on the server
        :param localfile: name of the downloaded file on the local filesystem,
            if `None` will have same name as remote file
        :return: true or false based on success or failure
        """
        if self.instance is None:
            raise FTPException("No FTP connection")
        try:
            cmd = f"RETR {remotefile}"
            if localfile is None:
                localfile = remotefile
            with open(localfile, "wb") as filepath:
                self.instance.retrbinary(cmd, filepath.write, 1024)
            notebook_file(localfile)
            return True
        except FileNotFoundError as e:
            self.logger.warning(str(e))
            return False
        except error_perm as e:
            self.logger.warning(str(e))
            os.unlink(localfile)
            return False
        except all_errors as e:
            raise FTPException from e

    @ftpcommand
    def cwd(self, dirname: str) -> bool:
        """Change working directory on the server

        :param dirname: name of the directory
        :return: true or false based on success or failure
        """
        self.instance.cwd(dirname)

    @ftpcommand
    def pwd(self) -> str:
        """Get current working directory on the server

        :return: current working directory name as a string
        """
        return self.instance.pwd()

    @ftpcommand
    def mkd(self, dirname: str) -> bool:
        """Create a new directory on the server

        :param dirname: name of the directory
        :return: true or false based on success or failure
        """
        self.instance.mkd(dirname)

    @ftpcommand
    def rmd(self, dirname: str) -> bool:
        """Remove directory on the server

        :param dirname: name of the directory
        :return: true or false based on success or failure
        """
        self.instance.rmd(dirname)

    @ftpcommand
    def list_files(self, dirname: str = "") -> list:
        """List files on the server directory

        :param dirname: name of the directory
        :return: list of files present in the server directory
        """
        try:
            return list(self.instance.mlsd(path=dirname))
        except all_errors:
            return list(self.instance.nlst())

    @ftpcommand
    def delete(self, filepath: str) -> bool:
        """Delete file on the server

        :param filepath: path to server file
        :return: true or false based on success or failure
        """
        self.instance.delete(filepath)

    @ftpcommand
    def rename(self, fromname: str, toname: str) -> bool:
        """Rename file on the server

        :param fromname: current name of the file
        :param toname: new name for the file
        :return: true or false based on success or failure
        """
        self.instance.rename(fromname, toname)

    @ftpcommand
    def send_command(self, command: str) -> bool:
        """Execute command on the server

        List of FTP commands:
        https://en.wikipedia.org/wiki/List_of_FTP_commands

        :param command: name of the command to send
        :return: true or false based on success or failure
        """
        return self.instance.sendcmd(command)

    @ftpcommand
    def file_size(self, filepath: str) -> int:
        """Return byte size of the file on the server

        :param filepath: path to server file
        :return: byte size as an int
        """
        self.set_binary_mode()
        return self.instance.size(filepath)

    @ftpcommand
    def abort(self) -> bool:
        """Abort a file transfer in progress

        :return: true or false based on success or failure
        """
        self.instance.abort()

    @ftpcommand
    def get_welcome_message(self) -> str:
        """Get server welcome message

        :return: welcome message as a string
        """
        return self.instance.getwelcome()

    @ftpcommand
    def set_debug_level(self, level: int = 0) -> bool:
        """Set debug level for the library

        :param level: integer value of debug level, defaults to 0
        :return: true or false based on success or failure

        0 - no debugging output
        1 - moderate amount of debugging
        2+ - higher amount of debugging
        """
        if level >= 0:
            self.instance.set_debuglevel(level)
        else:
            self.logger.warning("Valid debug levels are 0, 1 or 2+")

    def set_ascii_mode(self) -> bool:
        """Set transfer mode to ASCII

        :return: true or false based on success or failure
        """
        self.send_command("TYPE a")

    def set_binary_mode(self) -> bool:
        """Set transfer mode to BINARY

        :return: true or false based on success or failure
        """
        self.send_command("TYPE i")



import time
from typing import Optional
from PIL import ImageOps

from RPA.core.geometry import Region
from RPA.Desktop.keywords import LibraryContext, keyword, screen, HAS_RECOGNITION

if HAS_RECOGNITION:
    from RPA.recognition import ocr  # pylint: disable=no-name-in-module
else:
    ocr = None


def ensure_recognition():
    if not HAS_RECOGNITION:
        raise ValueError(
            "Keyword requires OCR features, please install the "
            "rpaframework-recognition package"
        )


class TextKeywords(LibraryContext):
    """Keywords for reading screen information and content."""

    @keyword
    def read_text(
        self,
        locator: Optional[str] = None,
        invert: bool = False,
        language: str = None,
        configuration: str = None,
    ):
        """Read text using OCR from the screen, or an area of the
        screen defined by the given locator.

        :param locator: Location of element to read text from
        :param invert:  Invert image colors, useful for reading white text
            on dark background
        :param language: 3-character ISO 639-2 language code of the text.
            This is passed directly to the pytesseract lib in the lang parameter.
            See https://tesseract-ocr.github.io/tessdoc/Command-Line-Usage.html#using-one-language
        :param configuration: Tesseract specific parameters like Page Segmentation
            Modes(psm) or OCR Engine Mode (oem). This is passed directly to the
            pytesseract lib in the config parameter.
            See https://tesseract-ocr.github.io/tessdoc/Command-Line-Usage.html

        Usage examples:

        .. code-block:: robotframework

            ${label_region}=  Find Element  image:label.png
            ${value_region}=  Move Region   ${label_region}  100  0
            ${text}=          Read Text     ${value_region}

        .. code-block:: python

            label_region = desktop.find_element("image:label.png")
            value_region = desktop.move_region(label_region, 100, 0)
            text = desktop.read_text(value_region)

        """  # noqa: E501
        ensure_recognition()

        if locator is not None:
            element = self.ctx.wait_for_element(locator)
            if not isinstance(element, Region):
                raise ValueError("Locator must resolve to a region")

            area = "element: %s" % element
            image = screen.grab(element)
        else:
            area = "screen"
            image = screen.grab()

        screen.log_image(image)

        if invert:
            image = ImageOps.invert(image)

        self.logger.info(
            "Reading text from %s (invert: %s, language: %s, configuration: %s)",
            area,
            invert or "Not set",
            language or "Not set",
            configuration or "Not set",
        )

        start_time = time.time()
        text = ocr.read(image, language, configuration)
        self.logger.info("Read text in %.2f seconds", time.time() - start_time)

        return text



# pylint: disable=too-many-lines
import atexit
import base64
import datetime
import json
import logging
import os
import platform
import shutil
import subprocess
import time
import traceback
import urllib.parse
import webbrowser
from collections import OrderedDict
from functools import partial
from itertools import product
from pathlib import Path
from typing import Any, Callable, Dict, List, Optional, Tuple, Union

from robot.libraries.BuiltIn import BuiltIn, RobotNotRunningError
from selenium import webdriver as selenium_webdriver
from selenium.common import WebDriverException
from selenium.common.exceptions import ElementClickInterceptedException
from selenium.webdriver import (
    ChromeOptions,
    EdgeOptions,
    FirefoxOptions as _FirefoxOptions,
    FirefoxProfile,
    IeOptions,
)
from selenium.webdriver.common.by import By
from selenium.webdriver.common.options import ArgOptions
from selenium.webdriver.remote.shadowroot import ShadowRoot
from selenium.webdriver.support import expected_conditions
from selenium.webdriver.support.ui import WebDriverWait
from SeleniumLibrary import EMBED, SeleniumLibrary, WebElement
from SeleniumLibrary.base import keyword
from SeleniumLibrary.errors import ElementNotFound
from SeleniumLibrary.keywords import (
    AlertKeywords,
    BrowserManagementKeywords as _BrowserManagementKeywords,
    ScreenshotKeywords,
    WindowKeywords as _WindowKeywords,
)
from SeleniumLibrary.keywords.webdrivertools import SeleniumOptions, WebDriverCreator
from SeleniumLibrary.locators import ElementFinder

from RPA.Browser.common import AUTO, auto_headless
from RPA.core import notebook
from RPA.core import webdriver as core_webdriver
from RPA.core.locators import BrowserLocator, LocatorsDatabase
from RPA.Robocorp.utils import get_output_dir


Element = Union[WebElement, ShadowRoot]
Locator = Union[Element, str]
AliasType = Union[str, int]
TimeoutType = Optional[Union[str, int, datetime.timedelta]]

OptionsType = Union[ArgOptions, str, Dict[str, Union[str, List, Dict]]]
ChromiumOptions = Union[ChromeOptions, EdgeOptions]


def html_table(header, rows):
    """Create HTML table that can be used for logging."""
    output = '<div class="doc"><table>'
    output += "<tr>" + "".join(f"<th>{name}</th>" for name in header) + "</tr>"
    for row in rows:
        output += "<tr>" + "".join(f"<td>{name}</td>" for name in row) + "</tr>"
    output += "</table></div>"
    return output


def ensure_scheme(url: str, default: Optional[str]) -> str:
    """Ensures that a URL has a scheme, such as `http` or `https`"""
    if not all([url, default]):
        return url  # nothing to do here in the absence of the URL or scheme

    parts = list(urllib.parse.urlsplit(url))
    if not parts[0]:
        parts[0] = str(default)
        url = urllib.parse.urlunsplit(parts)

    return url


class BrowserNotFoundError(ValueError):
    """Raised when browser can't be initialized."""


class RobocorpElementFinder(ElementFinder):
    """Customizes the element finding logic."""

    def _is_webelement(self, element: Element) -> bool:
        """Checks and accepts various web elements during finding."""
        # NOTE(cmin764): This will allow the finder to fully parse the locator and look
        #  for elements even under a shadow root.
        return isinstance(element, ShadowRoot) or super()._is_webelement(element)


class FirefoxOptions(_FirefoxOptions):
    """Wrapped Firefox options in order to fix behavior."""

    @property
    def binary_location(self) -> Optional[str]:
        # pylint: disable=protected-access
        return self.binary._start_cmd if self.binary else None


class BrowserManagementKeywords(_BrowserManagementKeywords):
    """Overridden keywords for browser management."""

    def __init__(self, ctx):
        super().__init__(ctx)
        self._default_scheme: Optional[str] = "https"

    @keyword
    def set_default_url_scheme(self, scheme: Optional[str]) -> None:
        """Sets the default `scheme` used for URLs without a defined
        value, such as `http` or `https`.

        The feature is disabled if the value is set to `None`.
        """
        self._default_scheme = scheme

    @keyword
    def go_to(self, url: str) -> None:
        url = ensure_scheme(url, default=self._default_scheme)
        super().go_to(url)

    go_to.__doc__ = _BrowserManagementKeywords.go_to.__doc__

    @keyword
    def open_browser(
        self,
        url: Optional[str] = None,
        browser: str = "firefox",
        alias: Optional[str] = None,
        remote_url: Union[bool, str] = False,
        desired_capabilities: Union[dict, None, str] = None,
        ff_profile_dir: Union[FirefoxProfile, str, None] = None,
        options: Optional[OptionsType] = None,
        service_log_path: Optional[str] = None,
        executable_path: Optional[str] = None,
    ) -> str:
        if url:
            url = ensure_scheme(url, default=self._default_scheme)
        if options:
            options: ArgOptions = self.ctx.normalize_options(options, browser=browser)
        return super().open_browser(
            url=url,
            browser=browser,
            alias=alias,
            remote_url=remote_url,
            desired_capabilities=desired_capabilities,
            ff_profile_dir=ff_profile_dir,
            options=options,
            service_log_path=service_log_path,
            executable_path=executable_path,
        )

    open_browser.__doc__ = _BrowserManagementKeywords.open_browser.__doc__


# pylint: disable=missing-class-docstring
class WindowKeywords(_WindowKeywords):
    # pylint: disable=arguments-differ
    @keyword
    def maximize_browser_window(self, *args, force: bool = False, **kwargs):
        """Maximizes current browser window.

        The window won't be maximized in headless mode since there's no way to know the
        screen size to set the window size to in the absence of an UI. Use the
        ``Set Window Size`` keyword with a specific side or set the `force` param to
        `True` if you still want to enforce this undefined behaviour.
        """
        if self.ctx.headless:
            decision = (
                "continuing with maximization" if force else "ignoring maximization"
            )
            self.ctx.logger.warning(
                "Attempting to maximize browser in headless mode without knowning the"
                " screen resolution. (%s)",
                decision,
            )
            if not force:
                return

        super().maximize_browser_window(*args, **kwargs)


class Selenium(SeleniumLibrary):
    # NOTE(cmin764): The docstring below will be appended (and not overridding) the
    #  docstring of the super-class.
    """= Auto closing browser =

    By default, the browser instances created during a task execution are closed
    at the end of the task. This can be prevented with the ``auto_close``
    parameter when *importing* the library.

    The value of the parameter needs to be set to ``False`` or any object evaluated as
    false (see `Boolean arguments`).
    """  # noqa: E501

    __doc__ = f"{SeleniumLibrary.__doc__}\n{__doc__}"

    ROBOT_LIBRARY_SCOPE = "GLOBAL"
    ROBOT_LIBRARY_DOC_FORMAT = "ROBOT"

    BROWSER_NAMES = {
        **WebDriverCreator.browser_names,
        "chromiumedge": WebDriverCreator.browser_names["edge"],
    }
    AVAILABLE_OPTIONS = {
        # Supporting options only for a specific range of browsers.
        "chrome": selenium_webdriver.ChromeOptions,
        "firefox": FirefoxOptions,
        "edge": selenium_webdriver.EdgeOptions,
        "chromiumedge": selenium_webdriver.EdgeOptions,
        "safari": selenium_webdriver.SafariOptions,
        "ie": selenium_webdriver.IeOptions,
    }
    AVAILABLE_SERVICES = {
        # Supporting services and their default webdriver binaries only for a specific
        #  range of browsers.
        "chrome": (selenium_webdriver.chrome.service.Service, "chromedriver"),
        "firefox": (selenium_webdriver.firefox.service.Service, "geckodriver"),
        "edge": (selenium_webdriver.edge.service.Service, "msedgedriver"),
        "chromiumedge": (selenium_webdriver.edge.service.Service, "msedgedriver"),
        "safari": (selenium_webdriver.safari.service.Service, "safaridriver"),
        "ie": (selenium_webdriver.ie.service.Service, "IEDriverServer"),
    }
    SUPPORTED_BROWSERS = dict(
        {name: name.capitalize() for name in AVAILABLE_SERVICES},
        **{"chromiumedge": "ChromiumEdge"},
    )
    # Both driver and browser lower-case names.
    CHROMIUM_BROWSERS = ["chrome", "edge", "chromiumedge", "msedge", "ie"]

    def __init__(self, *args, **kwargs):
        # We need to pop our kwargs before passing kwargs to SeleniumLibrary
        self.auto_close = kwargs.pop("auto_close", True)
        self.locators_path = kwargs.pop("locators_path", None)
        self.headless: Optional[bool] = None  # will be set once the browser starts

        # Parse user-given plugins
        plugins = kwargs.get("plugins", "")
        plugins = set(p for p in plugins.split(",") if p)

        # Add testability if requested
        if "use_testability" in args:
            args = [arg for arg in args if arg != "use_testability"]
            plugins.add("SeleniumTestability")

        # Refresh plugins list
        kwargs["plugins"] = ",".join(plugins)

        super().__init__(*args, **kwargs)
        self._element_finder = RobocorpElementFinder(self)

        # Add inherit/overridden library keywords.
        self.browser_management = BrowserManagementKeywords(self)
        window_kw = WindowKeywords(self)
        override_plugins = [self.browser_management, window_kw]
        self.add_library_components(override_plugins)

        self.logger = logging.getLogger(__name__)
        self.using_testability = bool("SeleniumTestability" in plugins)

        # Add support for locator aliases.
        self._element_finder.register("alias", self._find_by_alias, persist=True)

        # Embed screenshots in logs by default.
        if not notebook.IPYTHON_AVAILABLE:
            self._embedding_screenshots = True
            self._previous_screenshot_directory = self.set_screenshot_directory(EMBED)
        else:
            self._embedding_screenshots = False
            self._previous_screenshot_directory = None

        self.download_preferences = {}
        self._close_on_exit()

    def _close_on_exit(self):
        """Register cleanup function for leftover webdrivers & browsers on process
        exit.
        """

        def stop_drivers():
            if self.auto_close:
                self._quit_all_drivers()
            elif platform.system() == "Windows":
                # NOTE: On Windows, the webdriver executable keeps hanging and prevents
                #  "rcc" to close even when the Python process exits.
                self._quit_all_drivers(driver_only=True)

        atexit.register(stop_drivers)

    def _quit_all_drivers(self, driver_only: bool = False):
        # With `driver_only` on, we'll close just the drivers, but still leave the
        #  browser window open.
        connections = self._drivers._connections  # pylint: disable=protected-access
        for driver in connections:
            try:
                if driver_only:
                    # A `service.stop()` will hang here, so killing the process
                    #  directly is the only way.
                    driver.service.process.kill()
                else:
                    driver.quit()  # quits the browser as well
            except Exception as exc:  # pylint: disable=broad-except
                self.logger.warning("Encountered error during auto-close: %s", exc)

    @property
    def location(self) -> str:
        """Return browser location."""
        return self.get_location()

    def _find_by_alias(self, parent, criteria, tag, constraints):
        """Custom 'alias' locator that uses locators database."""
        locator = LocatorsDatabase.load_by_name(criteria, self.locators_path)

        if not isinstance(locator, BrowserLocator):
            raise ValueError(f"Not a browser locator: {criteria}")

        strategy = str(locator.strategy).lower()
        finder: Optional[str] = {
            "class": By.CLASS_NAME,
            "css": By.CSS_SELECTOR,
            "id": By.ID,
            "link": By.LINK_TEXT,
            "name": By.NAME,
            "tag": By.TAG_NAME,
            "xpath": By.XPATH,
        }.get(strategy)

        if not finder:
            raise ValueError(f"Unsupported locator strategy: {strategy}")

        # pylint: disable=protected-access
        return self._element_finder._filter_elements(
            parent.find_elements(finder, locator.value), tag, constraints
        )

    @keyword
    @auto_headless
    def open_available_browser(  # noqa: C901
        self,
        url: Optional[str] = None,
        use_profile: bool = False,
        headless: Union[bool, str] = AUTO,
        maximized: bool = False,
        browser_selection: Any = AUTO,
        alias: Optional[str] = None,
        profile_name: Optional[str] = None,
        profile_path: Optional[str] = None,
        preferences: Optional[dict] = None,
        proxy: str = None,
        user_agent: Optional[str] = None,
        download: Any = AUTO,
        options: Optional[OptionsType] = None,
        port: Optional[int] = None,
        sandbox: bool = False,
    ) -> AliasType:
        # pylint: disable=C0301
        """Attempts to open a browser on the user's device from a set of
        supported browsers. Automatically downloads a corresponding webdriver
        if none is already installed.

        Currently supported browsers: %s

        Optionally can be given a ``url`` as the first argument,
        to open the browser directly to the given page.

        Returns either a generated index or a custom ``alias`` for the
        browser instance. The returned value can be used to refer to that
        specific browser instance in other keywords.

        If the browser should start in a maximized window, this can be
        enabled with the argument ``maximized``, but is disabled by default.

        For certain applications it might also be required to force a
        certain user-agent string for Selenium, which can be overridden
        with the ``user_agent`` argument.

        WebDriver creation can be customized with ``options``. This accepts a class
        instance (e.g. ``ChromeOptions``), a string like
        `add_argument("--incognito");set_capability("acceptInsecureCerts", True)` or
        even a simple dictionary like:
        `{"arguments": ["--incognito"], "capabilities": {"acceptInsecureCerts": True}}`

        A custom ``port`` can be provided to start the browser webdriver without a
        randomly picked one. Make sure you provide every time a unique system-available
        local port if you plan to have multiple browsers being controlled in parallel.

        For incompatible web apps designed to work in Internet Explorer only, Edge can
        run in IE mode by simply setting `ie` in the ``browser_selection`` param.
        Robot example: https://github.com/robocorp/example-ie-mode-edge

        The ``sandbox`` argument can be used to enable the sandbox mode for the browser.
        By default browser is opened in `--no-sandbox` mode, but this started to cause
        issues on Chromium version 124. The `--no-sandbox` flag is set by default to
        preserve the older behavior.

        Example:

        | Open Available Browser | https://www.robocorp.com |
        | ${index}= | Open Available Browser | ${URL} | browser_selection=opera,firefox |
        | Open Available Browser | ${URL} | headless=${True} | alias=HeadlessBrowser |
        | Open Available Browser | ${URL} | options=add_argument("user-data-dir=path/to/data");add_argument("--incognito") |
        | Open Available Browser | ${URL} | port=${8888} |

        == Browser order ==

        The default order of supported browsers is based on the operating system
        and is as follows:

        | Platform    | Default order                         |
        | ``Windows`` | Chrome, Firefox, Edge         |
        | ``Linux``   | Chrome, Firefox, Edge         |
        | ``Darwin``  | Chrome, Firefox, Edge, Safari |

        The order can be overridden with a custom list by using the argument
        ``browser_selection``. The argument can be either a comma-separated
        string or a list object.

        Example:

        | Open Available Browser | ${URL} | browser_selection=ie |

        == Webdriver download ==

        The library can (if requested) automatically download webdrivers
        for all the supported browsers. This can be controlled with the argument
        ``download``.

        If the value is ``False``, it will only attempt to start
        webdrivers found from the system PATH.

        If the value is ``True``, it will download a webdriver that matches
        the current browser.

        By default the argument has the value ``AUTO``, which means it
        first attempts to use webdrivers found in PATH and if that fails
        forces a webdriver download.

        == Opening process ==

        1. Parse list of preferred browser order. If not given, use values
           from above table.

        2. Loop through listed browsers:

            a. Set the webdriver options for the browser.

            b. Download webdriver (if requested).

            c. Attempt to launch the webdriver and stop the loop if successful.

        3. Return index/alias if webdriver was created, or raise an exception
           if no browsers were successfully opened.

        == Headless mode ==

        If required, the browser can also run `headless`, which means that
        it does not create a visible window. Generally a headless browser is
        slightly faster, but might not support all features a normal browser does.

        One typical use-case for headless mode is in cloud containers,
        where there is no display available. It also prevents manual interaction
        with the browser, which can be either a benefit or a drawback depending on
        the context.

        It can be explicitly enabled or disabled with the argument ``headless``.
        By default, it will be disabled, unless it detects that it is running
        in a Linux environment without a display, e.g. a container or if the
        `RPA_HEADLESS_MODE` env var is set to a number different than `0`.

        == Chromium options ==

        Some features are currently available only for Chromium-based browsers.
        This includes using an existing user profile. By default Selenium
        uses a new profile for each session, but it can use an existing
        one by enabling the ``use_profile`` argument.

        If a custom profile is stored somewhere outside of the default location,
        the path to the profiles directory and the name of the profile can
        be controlled with ``profile_path`` and ``profile_name`` respectively. Keep in
        mind that the ``profile_path`` for the Chrome browser for e.g. ends usually
        with "Chrome", "User Data" or "google-chrome" (based on platform) and the
        ``profile_name`` is a directory relative to ``profile_path``, usually named
        "Profile 1", "Profile 2" etc. (and not as your visible name in the Chrome
        browser). Similar behavior is observed with Edge as well.

        Example:

        | Open Available Browser | https://www.robocorp.com | use_profile=${True} |
        | Open Available Browser | https://www.robocorp.com | use_profile=${True} | profile_name=Default |
        | Open Available Browser | https://www.robocorp.com | use_profile=${True} | profile_name=Profile 2 |
        | Open Available Browser | https://www.robocorp.com | use_profile=${True} | profile_name=Profile 1 | profile_path=path/to/custom/user_data_dir |

        Profile preferences can be further overridden with the ``preferences``
        argument by giving a dictionary of key/value pairs.

        Chromium-based browsers can additionally connect through a ``proxy``, which
        should be given as either a local or remote address.
        """  # noqa: E501
        # pylint: disable=redefined-argument-from-local
        browsers = self._arg_browser_selection(browser_selection)
        downloads = self._arg_download(download)

        attempts = []
        index_or_alias = None
        browser_found = True
        current_browser = None
        self.headless: bool = headless  # it's resolved through the decorator

        # Try all browsers in preferred order
        for browser, download in product(browsers, downloads):
            if current_browser != browser:
                current_browser = browser
                browser_found = True
            elif not browser_found:
                continue
            try:
                kwargs, arguments, browser_version = self._get_driver_args(
                    browser,
                    headless,
                    maximized,
                    use_profile,
                    profile_name,
                    profile_path,
                    preferences,
                    proxy,
                    user_agent,
                    options,
                    port,
                    url,
                    sandbox,
                )
                if not browser_version:
                    attempts.append(
                        (browser, download, "Could not get browser version")
                    )
                    self.logger.info(
                        "Could not detect version of the %s browser", browser
                    )
                    browser_found = False
                    continue
                self.logger.debug(
                    "Creating webdriver for %r (headless: %s, download: %s)",
                    browser,
                    headless,
                    download,
                )
                index_or_alias = self._create_webdriver(
                    browser, alias, download, **kwargs
                )
                attempts.append((browser, download, ""))
                self.logger.info(
                    "Created %s browser (version %s) with arguments: %s",
                    browser,
                    browser_version,
                    " ".join(arguments),
                )
                break
            except Exception as error:  # pylint: disable=broad-except
                attempts.append((browser, download, error))
                self.logger.debug(traceback.format_exc())

        # Log table of all attempted combinations
        table_headers = ["Browser", "Download", "Error"]
        try:
            table = html_table(header=table_headers, rows=attempts)
            BuiltIn().log("<p>Attempted combinations:</p>" + table, html=True)
        except RobotNotRunningError:
            pass

        # No webdriver was started
        if index_or_alias is None:
            errors = OrderedDict((browser, error) for browser, _, error in attempts)

            msg = "Failed to start a browser:\n"
            for browser, error in errors.items():
                msg += f"- {browser}: {error}\n"

            notebook.notebook_table(attempts, columns=table_headers)
            raise BrowserNotFoundError(msg)

        if url is not None:
            self.go_to(url)

        return index_or_alias

    open_available_browser.__doc__ %= ", ".join(SUPPORTED_BROWSERS.values())

    def _arg_browser_selection(
        self, browser_selection: Union[str, List[str]]
    ) -> List[str]:
        """Parse argument for browser selection."""
        if str(browser_selection).strip().lower() == AUTO.lower():
            order = core_webdriver.get_browser_order()
        else:
            order = (
                browser_selection
                if isinstance(browser_selection, list)
                else [browser.strip() for browser in browser_selection.split(",")]
            )
        return order

    def _arg_download(self, download: Any) -> List:
        """Parse argument for webdriver download."""
        if str(download).strip().lower() == AUTO.lower():
            return [False, True]
        else:
            return [bool(download)]

    def _set_experimental_options(
        self, browser_lower: str, options: ArgOptions, preferences: Optional[dict]
    ):
        default_preferences = {
            "safebrowsing.enabled": True,
            "credentials_enable_service": False,
            "profile.password_manager_enabled": False,
        }
        options.add_experimental_option(
            "prefs",
            {
                **default_preferences,
                **self.download_preferences.get(browser_lower, {}),
                **(preferences or {}),
            },
        )
        options.add_experimental_option(
            "excludeSwitches", ["enable-logging", "enable-automation"]
        )
        if not self.auto_close:
            # Leave the browser window open if auto-closing is disabled.
            options.add_experimental_option("detach", True)

    def _set_chromium_options(
        self,
        browser_lower: str,
        kwargs: dict,
        options: ArgOptions,
        use_profile: bool = False,
        profile_name: Optional[str] = None,
        profile_path: Optional[str] = None,
        preferences: Optional[dict] = None,
        proxy: str = None,
        sandbox: bool = False,
    ):
        if proxy:
            options.add_argument(f"--proxy-server={proxy}")
        options.add_argument("--disable-dev-shm-usage")
        options.add_argument("--disable-web-security")
        options.add_argument("--allow-running-insecure-content")
        if not sandbox:
            options.add_argument("--no-sandbox")

        if browser_lower != "ie":
            # While Edge supports experimental options, Edge in IE mode doesn't.
            self._set_experimental_options(browser_lower, options, preferences)
        if use_profile:
            # NOTE(cmin764; 14 Sep 2023): This doesn't currently work with IE due to a
            #  bug in the webdriver.
            # Issue: https://github.com/MicrosoftEdge/EdgeWebDriver/issues/29
            self._set_user_profile(browser_lower, options, profile_path, profile_name)
        if self.logger.isEnabledFor(logging.DEBUG):
            # Deprecated params, but no worries as they get popped then bundled in a
            #  `Service` instance inside of the `self._create_webdriver` method.
            kwargs["service_log_path"] = "chromiumdriver.log"
            kwargs["service_args"] = ["--verbose"]

    def _set_ie_options(self, options: IeOptions, *, url: Optional[str]):
        binary_location = getattr(options, "binary_location", None)
        if binary_location:
            options.edge_executable_path = binary_location
        # An invalid default URL might make the automation freeze.
        options.initial_browser_url = (
            options.initial_browser_url or url or "https://robocorp.com/"
        )

    def _set_firefox_options(
        self, options: FirefoxOptions, *, preferences: Optional[dict]
    ):
        # Set the custom download directory previously configured with
        #  `set_download_directory(...)`.
        prefs = {
            **self.download_preferences.get("firefox", {}),
            **(preferences or {}),
        }
        for name, value in prefs.items():
            options.set_preference(name, value)

    def _set_option(
        self, name: str, values: Union[str, List, Dict], *, method: Callable
    ):
        if name == "arguments":
            if isinstance(values, str):
                values = [value.strip() for value in values.split(",")]
            for value in values:
                self.logger.debug("Setting argument: %s", value)
                method(value)
        elif name == "capabilities":
            if isinstance(values, str):
                values, _values = {}, values
                for item in _values.split(","):
                    key, val = item.strip().split(":")
                    values[key.strip()] = val.strip()
            for key, val in values.items():
                self.logger.debug("Setting capability: %s=%s", key, val)
                method(key, val)
        elif name == "binary_location":
            self.logger.debug("Setting binary location: %s", values)
            method(values)

    @staticmethod
    def _set_options_from_env(options_obj: ArgOptions):
        """Add default options from the environment variables when an explicit value
        isn't set already.
        """
        binary_location = os.getenv("RPA_SELENIUM_BINARY_LOCATION")
        if binary_location and not getattr(options_obj, "binary_location"):
            options_obj.binary_location = binary_location

    def normalize_options(
        self, options: Optional[OptionsType], *, browser: str
    ) -> ArgOptions:
        """Normalize provided `options` to a `<Browser>Options` instance."""
        browser_lower = browser.lower()

        # String or object based provided options, solved by the wrapped library.
        if isinstance(options, (ArgOptions, str)):
            options_obj = SeleniumOptions().create(
                self.BROWSER_NAMES[browser_lower], options
            )
            self._set_options_from_env(options_obj)
            return options_obj

        BrowserOptions = self.AVAILABLE_OPTIONS.get(browser_lower)
        if not BrowserOptions:
            raise ValueError(
                f"{browser!r} browser options not provided as a string/object aren't"
                f" supported"
            )

        options_obj = BrowserOptions()
        if isinstance(options, dict):
            option_method_map = {
                "arguments": options_obj.add_argument,
                "capabilities": options_obj.set_capability,
                "binary_location": lambda path: setattr(
                    options_obj, "binary_location", path
                ),
            }
            for name, values in options.items():
                if name not in option_method_map:
                    raise TypeError(
                        f"Option type {name!r} not supported, choose from"
                        f" {list(option_method_map)} or try providing them as string"
                        " or object"
                    )
                method = option_method_map[name]
                self._set_option(name, values, method=method)

        self._set_options_from_env(options_obj)
        return options_obj

    def _get_driver_args(  # noqa: C901
        self,
        browser: str,
        headless: bool = False,
        maximized: bool = False,
        use_profile: bool = False,
        profile_name: Optional[str] = None,
        profile_path: Optional[str] = None,
        preferences: Optional[dict] = None,
        proxy: str = None,
        user_agent: Optional[str] = None,
        options: Optional[OptionsType] = None,
        port: Optional[int] = None,
        url: Optional[str] = None,
        sandbox: bool = False,
    ) -> Tuple[dict, Any, str]:
        """Get browser and webdriver arguments for given options."""
        browser_lower = browser.lower()
        if browser_lower not in self.AVAILABLE_OPTIONS:
            return {}, [], None

        options: ArgOptions = self.normalize_options(options, browser=browser)
        if headless:
            self._set_headless_options(browser_lower, options)
        if maximized and not headless:
            options.add_argument("--start-maximized")
        if user_agent:
            options.add_argument(f"user-agent={user_agent}")

        kwargs = {}
        if port:
            # Deprecated kwarg which will be transferred into a service instance.
            kwargs["port"] = int(port)
        if browser_lower in self.CHROMIUM_BROWSERS:
            self._set_chromium_options(
                browser_lower,
                kwargs,
                options,
                use_profile=use_profile,
                profile_name=profile_name,
                profile_path=profile_path,
                preferences=preferences,
                proxy=proxy,
                sandbox=sandbox,
            )
            if browser_lower == "ie":
                self._set_ie_options(options, url=url)
        elif browser_lower == "firefox":
            self._set_firefox_options(options, preferences=preferences)
        if self.download_preferences and browser_lower not in self.download_preferences:
            self.logger.warning(
                "Custom download directory not supported with %r!", browser
            )
        if use_profile and browser_lower not in self.CHROMIUM_BROWSERS:
            self.logger.warning(
                "Profiles are supported with Chromium-based browsers only!"
            )

        path = getattr(options, "binary_location", None)
        if path:
            self.logger.warning(
                f"The custom provided browser ({path}) might be "
                "incompatible with the default downloaded webdriver. Use "
                "``Open Browser`` with these `options` and a compatible "
                "`executable_path` if running into issues."
            )

        kwargs["options"] = options  # legitimate webdriver kwarg separate from service
        browser_version = self._get_browser_version(browser, options=options)
        return kwargs, options.arguments, browser_version

    def _set_headless_options(self, browser_lower: str, options: ArgOptions) -> None:
        """Set headless mode for the browser, if possible.

        ``browser_lower`` is the lower-case string name of the browser.
        ``options`` are the browser options to be updated to match the headless config.
        """
        if browser_lower == "safari":
            self.logger.warning(
                "Safari does not support headless mode."
                " (https://github.com/SeleniumHQ/selenium/issues/5985)"
            )
            return

        # NOTE(cmin764): `options.headless` will be removed in Selenium 4.10.0
        #  (https://www.selenium.dev/blog/2023/headless-is-going-away/)
        headless_args = {
            "chrome": "--headless=new",
            "edge": "--headless=new",
            "chromiumedge": "--headless=new",
            "firefox": "-headless",
        }
        headless_arg = headless_args.get(browser_lower)
        if headless_arg:
            options.add_argument(headless_arg)
        else:
            self.logger.warning(
                "Headless is supported only with: %s", ", ".join(headless_args)
            )
        options.add_argument("--disable-gpu")

        if browser_lower in self.CHROMIUM_BROWSERS:
            options.add_argument("--window-size=1440,900")

    def _set_user_profile(
        self,
        browser_lower: str,
        options: ArgOptions,
        profile_path: Optional[str] = None,
        profile_name: Optional[str] = None,
    ) -> None:
        """Set user profile configuration into the browser options.

        In the absence of an explicit `profile_path` an `RPA_CHROME_USER_PROFILE_DIR`
        environment variable is required to point into the user profile directory
        instead. Optionally, if a `profile_name` is passed, then we'll be using that
        instead of the default one.
        """
        data_dir = profile_path or os.getenv("RPA_CHROME_USER_PROFILE_DIR")
        system = platform.system()
        home = Path.home()

        if not data_dir:
            data_dirs = {
                "Windows": {
                    "chrome": home
                    / "AppData"
                    / "Local"
                    / "Google"
                    / "Chrome"
                    / "User Data",
                    "edge": home
                    / "AppData"
                    / "Local"
                    / "Microsoft"
                    / "Edge"
                    / "User Data",
                },
                "Linux": {
                    "chrome": home / ".config" / "google-chrome",
                    "edge": home / ".config" / "edge",
                },
                "Darwin": {
                    "chrome": home
                    / "Library"
                    / "Application Support"
                    / "Google"
                    / "Chrome",
                    "edge": home / "Library" / "Application Support" / "Microsoft Edge",
                },
            }
            for value in data_dirs.values():
                value["chromiumedge"] = value["edge"]
            data_dir = data_dirs.get(system, {}).get(browser_lower)
        if not data_dir:
            self.logger.warning(
                "Unable to resolve profile directory on %r for: %s",
                system,
                browser_lower,
            )
            return

        if not Path(data_dir).exists():
            self.logger.warning("Given profile directory does not exist: %s", data_dir)

        options.add_argument("--enable-local-sync-backend")
        options.add_argument(f"--local-sync-backend-dir={data_dir}")
        options.add_argument(f"--user-data-dir={data_dir}")
        if profile_name is not None:
            options.add_argument(f"--profile-directory={profile_name}")

    def _get_service_class(self, lower_browser: str) -> type:
        Service, default_webdriver = self.AVAILABLE_SERVICES[lower_browser]

        class BrowserService(Service):
            """Custom service class wrapping the picked browser's one."""

            # pylint: disable=no-self-argument
            def __init__(
                this, *args, executable_path: str = default_webdriver, **kwargs
            ):
                # NOTE(cmin764): Starting with Selenium 4.9.1, we have to block their
                #  `SeleniumManager` from early stage, otherwise it will be activated
                #  when the `WebDriver` class itself if instantiated with a null
                #  executable path.
                resolved_path = (
                    shutil.which(executable_path) if executable_path else None
                )
                if not resolved_path:
                    raise WebDriverException(
                        f"Webdriver executable {executable_path!r} is not in PATH"
                        " (with disabled Selenium Manager)"
                    )

                this._log_webdriver_version(resolved_path)
                super().__init__(*args, executable_path=resolved_path, **kwargs)

            @staticmethod
            def _log_webdriver_version(path: str):
                args = [path, "--version"]
                try:
                    result = subprocess.run(
                        args, check=True, capture_output=True, text=True
                    )
                # pylint: disable=broad-except
                except Exception as exc:
                    self.logger.warning(
                        "Webdriver version couldn't be read due to: %s", exc
                    )
                else:
                    self.logger.info(
                        "Webdriver version taken into use: %s", result.stdout
                    )

            # pylint: disable=no-self-argument
            def __del__(this) -> None:
                # With auto-close disabled, we shouldn't call the object's cleanup
                #  method, as this will automatically stop the webdriver service, which
                #  implies a browser shutdown command, which closes the browser.
                if self.auto_close:
                    super().__del__()

        return BrowserService

    def _get_browser_version(self, browser: str, *, options: ArgOptions):
        cap_browser = browser[0].upper() + browser[1:]
        binary_location = getattr(options, "binary_location", None)
        try:
            version = core_webdriver.get_browser_version(
                cap_browser, path=binary_location
            )
        except Exception:  # pylint: disable=broad-except
            return None
        return version

    def _create_webdriver(
        self, browser: str, alias: Optional[str], download: bool, **kwargs
    ) -> AliasType:
        """Create a webdriver instance with given options.

        If webdriver download is requested, a cached version will be used if exists.
        """

        def _create_driver(path: Optional[str] = None) -> AliasType:
            # Prepare webdriver's service instance keyword arguments.
            service_kwargs = {
                # Deprecated params if passed directly to the `WebDriver` class.
                "service_args": None,
                "service_log_path": None,
                "port": 0,
            }
            for name, default in service_kwargs.items():
                service_kwargs[name] = kwargs.pop(name, default)
            service_kwargs["log_path"] = service_kwargs.pop("service_log_path")
            if path:
                service_kwargs["executable_path"] = path

            # Instantiate the right service to be passed during the webdriver creation.
            lower_browser = browser.lower()
            if lower_browser == "safari":
                service_kwargs.pop("log_path")  # not supported at all
            elif lower_browser == "ie":
                service_kwargs["log_file"] = service_kwargs.pop("log_path")
            BrowserService = self._get_service_class(lower_browser)
            service = BrowserService(**service_kwargs)

            # Capitalize browser name just to ensure it works if passed as lower case.
            # NOTE: But don't break a browser name like "ChromiumEdge".
            cap_browser = browser[0].upper() + browser[1:]
            kwargs["service"] = service
            return self.browser_management.create_webdriver(
                cap_browser, alias, **kwargs
            )

        # No download requested.
        if not download:
            return _create_driver()

        # Download web driver. (caching is tackled internally)
        driver_path = core_webdriver.download(browser)
        self.logger.debug(f"web driver path: {driver_path}")
        return _create_driver(path=driver_path)

    @keyword
    def open_chrome_browser(
        self,
        url: str,
        use_profile: bool = False,
        headless: Union[bool, str] = AUTO,
        maximized: bool = False,
        alias: Optional[str] = None,
        profile_name: Optional[str] = None,
        profile_path: Optional[str] = None,
        preferences: Optional[dict] = None,
        proxy: str = None,
        user_agent: Optional[str] = None,
        sandbox: bool = False,
    ) -> AliasType:
        """Opens a Chrome browser.

        See ``Open Available Browser`` for a full descriptions of the arguments.
        """
        return self.open_available_browser(
            url,
            alias=alias,
            headless=headless,
            maximized=maximized,
            use_profile=use_profile,
            browser_selection="Chrome",
            profile_name=profile_name,
            profile_path=profile_path,
            preferences=preferences,
            proxy=proxy,
            user_agent=user_agent,
            sandbox=sandbox,
        )

    @keyword
    def attach_chrome_browser(
        self, port: int, alias: Optional[str] = None
    ) -> AliasType:
        """Attach to an existing instance of Chrome browser.

        Requires that the browser was started with the command line
        option ``--remote-debugging-port=<port>``, where port is any
        4-digit number not being used by other applications.

        *Note.* The first Chrome instance on the system needs to be
        started with this command line option or this won't have an effect.

        That port can then be used to connect using this keyword.

        Example:

        | Attach Chrome Browser | port=9222 |
        """
        options = ChromeOptions()
        options.add_experimental_option("debuggerAddress", f"localhost:{port:d}")
        create = partial(self._create_webdriver, "Chrome", alias, options=options)

        try:
            return create(download=False)
        except Exception:  # pylint: disable=broad-except
            self.logger.debug(traceback.format_exc())
        return create(download=True)

    @keyword
    def open_headless_chrome_browser(self, url: str) -> AliasType:
        """Opens the Chrome browser in headless mode.

        ``url`` URL to open

        Example:

        | ${idx} = | Open Headless Chrome Browser | https://www.google.com |
        """
        return self.open_chrome_browser(url, headless=True)

    @keyword
    def screenshot(
        self,
        locator: Optional[Locator] = None,
        filename: Optional[str] = "",
    ) -> Optional[str]:
        # pylint: disable=C0301, W0212
        """Capture page and/or element screenshot.

        ``locator`` if defined, take element screenshot, if not takes page screenshot

        ``filename`` filename for the screenshot, by default creates file `screenshot-<timestamp>-(element|page).png`
        if set to `None` then file is not saved at all

        Example:

        | Screenshot | locator=//img[@alt="Google"] | filename=locator.png              | # element screenshot, defined filename            |
        | Screenshot | filename=page.png            |                                   | # page screenshot, defined filename               |
        | Screenshot | filename=${NONE}             |                                   | # page screenshot, NO file will be created        |
        | Screenshot |                              |                                   | # page screenshot, default filename               |
        | Screenshot | locator=//img[@alt="Google"] |                                   | # element screenshot, default filename            |
        | Screenshot | locator=//img[@alt="Google"] | filename=${CURDIR}/subdir/loc.png | # element screenshot, create dirs if not existing |
        """  # noqa: E501
        screenshot_keywords = ScreenshotKeywords(self)
        default_filename_prefix = f"screenshot-{int(time.time())}"

        def __save_base64_screenshot_to_file(base64_string, fname):
            path = screenshot_keywords._get_screenshot_path(fname)
            screenshot_keywords._create_directory(path)
            with open(fname, "wb") as fh:
                fh.write(base64.b64decode(base64_string))
                self.logger.info("Screenshot saved to file: %s", fname)

        if locator:
            element = screenshot_keywords.find_element(locator)
            ss_data = element.screenshot_as_base64
            screenshot_keywords._embed_to_log_as_base64(ss_data, 400)
            if filename is not None:
                filename = filename or os.path.join(
                    os.curdir, f"{default_filename_prefix}-element.png"
                )
                __save_base64_screenshot_to_file(ss_data, filename)
                notebook.notebook_image(filename)
        else:
            ss_data = self.driver.get_screenshot_as_base64()
            screenshot_keywords._embed_to_log_as_base64(ss_data, 800)
            if filename is not None:
                filename = filename or os.path.join(
                    os.curdir, f"{default_filename_prefix}-page.png"
                )
                __save_base64_screenshot_to_file(ss_data, filename)
                notebook.notebook_image(filename)

        return filename

    @keyword
    def click_element_when_visible(
        self,
        locator: Locator,
        modifier: Optional[str] = None,
        action_chain: bool = False,
    ) -> None:
        """Click element identified by ``locator``, once it becomes visible.

        ``locator`` element locator

        ``modifier`` press given keys while clicking the element, e.g. CTRL

        ``action_chain`` store action in Selenium ActionChain queue

        Example:

        | Click Element When Visible | q |
        | Click Element When Visible | id:button | CTRL+ALT |
        | Click Element When Visible | action_chain=True |
        """
        self.wait_until_element_is_visible(locator)
        self.click_element(locator, modifier, action_chain)

    @keyword
    def click_button_when_visible(
        self, locator: Locator, modifier: Optional[str] = None
    ) -> None:
        """Click button identified by ``locator``, once it becomes visible.

        ``locator`` element locator

        ``modifier`` press given keys while clicking the element, e.g. CTRL

        Example:

        | Click Button When Visible  | //button[@class="mybutton"] |
        """
        self.wait_until_element_is_visible(locator)
        self.click_button(locator, modifier)

    # Alias for backwards compatibility
    wait_and_click_button = click_button_when_visible

    @keyword
    def click_element_if_visible(self, locator: Locator) -> None:
        """Click element if it is visible

        ``locator`` element locator

        Example:

        | Click Element If Visible | //button[@class="mybutton"] |
        """
        visible = self.is_element_visible(locator)
        if visible:
            self.click_element(locator)

    @keyword
    def input_text_when_element_is_visible(self, locator: Locator, text: str) -> None:
        # pylint: disable=C0301
        """Input text into locator after it has become visible.

        ``locator`` element locator

        ``text`` insert text to locator

        Example:

        | Input Text When Element Is Visible | //input[@id="freetext"]  | my feedback |
        """  # noqa: E501
        self.wait_until_element_is_visible(locator)
        self.input_text(locator, text)

    @keyword
    def is_element_enabled(self, locator: Locator, missing_ok: bool = True) -> bool:
        """Is element enabled

        ``locator`` element locator
        ``missing_ok`` default True, set to False if keyword should
        Fail if element does not exist

        Example:

        | ${res} | Is Element Enabled | input.field1 |
        """
        return self._run_should_keyword_and_return_status(
            self.element_should_be_enabled,
            locator,
            missing_ok=missing_ok,
        )

    @keyword
    def is_element_visible(self, locator: Locator, missing_ok: bool = True) -> bool:
        """Is element visible

        ``locator`` element locator
        ``missing_ok`` default True, set to False if keyword should
        Fail if element does not exist

        Example:

        | ${res} | Is Element Visible | id:confirmation |
        """
        return self._run_should_keyword_and_return_status(
            self.element_should_be_visible,
            locator,
            missing_ok=missing_ok,
        )

    @keyword
    def is_element_disabled(self, locator: Locator, missing_ok: bool = True) -> bool:
        """Is element disabled

        ``locator`` element locator
        ``missing_ok`` default True, set to False if keyword should
        Fail if element does not exist

        Example:

        | ${res} | Is Element Disabled | //input[@type="submit"] |
        """
        return self._run_should_keyword_and_return_status(
            self.element_should_be_disabled,
            locator,
            missing_ok=missing_ok,
        )

    @keyword
    def is_element_focused(self, locator: Locator, missing_ok: bool = True) -> bool:
        """Is element focused

        ``locator`` element locator
        ``missing_ok`` default True, set to False if keyword should
        Fail if element does not exist

        Example:

        | ${res} | Is Element Focused | //input[@id="freetext"] |
        """
        return self._run_should_keyword_and_return_status(
            self.element_should_be_focused,
            locator,
            missing_ok=missing_ok,
        )

    @keyword
    def is_element_attribute_equal_to(
        self, locator: Locator, attribute: str, expected: str
    ) -> bool:
        """Is element attribute equal to expected value

        ``locator`` element locator

        ``attribute`` element attribute to check for

        ``expected`` is attribute value equal to this

        Example:

        | ${res} | Is Element Attribute Equal To | h1 | id | main |
        """
        return self._run_should_keyword_and_return_status(
            self.element_attribute_value_should_be, locator, attribute, expected
        )

    @keyword
    def is_alert_present(self, text: str = None, action: str = "ACCEPT") -> bool:
        """Is alert box present, which can be identified with text
        and action can also be done which by default is ACCEPT.

        Other possible actions are DISMISS and LEAVE.

        ``text`` check if alert text is matching to this, if `None`
        will check if alert is present at all

        ``action`` possible action if alert is present, default ACCEPT

        Example:

        | ${res} | Is Alert Present | alert message |
        """
        return self._run_should_keyword_and_return_status(
            self.alert_should_be_present, text, action
        )

    @keyword
    def does_alert_contain(self, text: str = None, timeout: TimeoutType = None) -> bool:
        # pylint: disable=W0212
        """Does alert contain text.

        ``text`` check if alert includes text, will raise ValueError is text
        does not exist

        Example:

        | ${res} | Does Alert Contain | alert message |
        """
        alert_keywords = AlertKeywords(self)
        alert = alert_keywords._wait_alert(timeout)
        if text in alert.text:
            return True
        else:
            raise ValueError(f"Alert did not contain text {text!r}")

    @keyword
    def does_alert_not_contain(
        self, text: str = None, timeout: TimeoutType = None
    ) -> bool:
        # pylint: disable=W0212
        """Does alert not contain text.

        ``text`` check that alert does not include text, will raise ValueError if text
        does exist

        Example:

        | ${res} | Does Alert Not Contain | unexpected message |
        """
        alert_keywords = AlertKeywords(self)
        alert = alert_keywords._wait_alert(timeout)

        if alert and text not in alert.text:
            return True
        else:
            raise ValueError(f"Alert did contain text {text!r}")

    @keyword
    def is_checkbox_selected(self, locator: Locator) -> bool:
        """Is checkbox selected

        ``locator`` element locator

        Example:

        | ${res} |  Is Checkbox Selected  | id:taxes-paid |
        """
        return self._run_should_keyword_and_return_status(
            self.checkbox_should_be_selected, locator
        )

    @keyword
    def does_frame_contain(self, locator: Locator, text: str) -> bool:
        """Does frame contain expected text

        ``locator`` locator of the frame to check

        ``text`` does frame contain this text

        Example:

        | ${res} | Does Frame Contain | id:myframe | secret |
        """
        return self._run_should_keyword_and_return_status(
            self.frame_should_contain, locator, text
        )

    @keyword
    def does_element_contain(
        self, locator: Locator, expected: str, ignore_case: bool = False
    ) -> bool:
        # pylint: disable=C0301
        """Does element contain expected text

        ``locator`` element locator

        ``expected`` expected element text

        ``ignore_case`` should check be case insensitive, default `False`

        Example:

        | ${res} | Does Element Contain | id:spec | specification complete | ignore_case=True |
        """  # noqa: E501
        return self._run_should_keyword_and_return_status(
            self.element_should_contain,
            locator=locator,
            expected=expected,
            ignore_case=ignore_case,
        )

    @keyword
    def is_element_text(
        self, locator: Locator, expected: str, ignore_case: bool = False
    ) -> bool:
        """Is element text expected

        ``locator`` element locator

        ``expected`` expected element text

        ``ignore_case`` should check be case insensitive, default `False`

        Example:

        | ${res} | Is Element Text | id:name | john doe |
        | ${res} | Is Element Text | id:name | john doe | ignore_case=True |
        """
        return self._run_should_keyword_and_return_status(
            self.element_text_should_be,
            locator=locator,
            expected=expected,
            ignore_case=ignore_case,
        )

    @keyword
    def is_list_selection(self, locator: Locator, *expected: str) -> bool:
        """Is list selected with expected values

        ``locator`` element locator

        ``expected`` expected selected options

        Example:

        | ${res} | Is List Selection | id:cars | Ford |
        """
        return self._run_should_keyword_and_return_status(
            self.list_selection_should_be, locator, *expected
        )

    @keyword
    def is_list_selected(self, locator: Locator) -> bool:
        """Is any option selected in the

        ``locator`` element locator

        Example:

        | ${res} | Is List Selected | id:cars |
        """
        self.logger.info("Will return if anything is selected on the list")
        return not self._run_should_keyword_and_return_status(
            self.list_should_have_no_selections, locator
        )

    @keyword
    def is_location(self, url: str) -> bool:
        """Is current URL expected url

        ``url`` expected current URL

        Example:

        | Open Available Browser | https://www.robocorp.com |
        | ${res} | Is Location | https://www.robocorp.com |
        """
        return self._run_should_keyword_and_return_status(self.location_should_be, url)

    @keyword
    def does_location_contain(self, expected: str) -> bool:
        """Does current URL contain expected

        ``expected`` URL should contain this

        Example:

        | Open Available Browser | https://robocorp.com |
        | ${res} | Does Location Contain | robocorp |
        """
        return self._run_should_keyword_and_return_status(
            self.location_should_contain, expected
        )

    @keyword
    def does_page_contain(self, text: str) -> bool:
        """Does page contain expected text

        ``text`` page should contain this

        Example:

        | Open Available Browser | https://google.com |
        | ${res} | Does Page Contain | Gmail |
        """
        return self._run_should_keyword_and_return_status(
            self.page_should_contain, text
        )

    @keyword
    def does_page_contain_button(self, locator: Locator) -> bool:
        """Does page contain expected button

        ``locator`` element locator

        Example:

        | ${res} | Does Page Contain Button | search-button |
        """
        return self._run_should_keyword_and_return_status(
            self.page_should_contain_button, locator
        )

    @keyword
    def does_page_contain_checkbox(self, locator: Locator) -> bool:
        """Does page contain expected checkbox

        ``locator`` element locator

        Example:

        | ${res} | Does Page Contain Checkbox | random-selection |
        """
        return self._run_should_keyword_and_return_status(
            self.page_should_contain_checkbox, locator
        )

    @keyword
    def does_page_contain_element(self, locator: Locator, count: int = None) -> bool:
        """Does page contain expected element

        ``locator`` element locator

        ``count`` how many times element is expected to appear on page
        by default one or more

        Example:

        | ${res} | Does Page Contain Element | textarea |
        | ${res} | Does Page Contain Element | button | count=4 |
        """
        return self._run_should_keyword_and_return_status(
            self.page_should_contain_element, locator=locator, limit=count
        )

    @keyword
    def does_page_contain_image(self, locator: Locator) -> bool:
        """Does page contain expected image

        ``locator`` element locator

        Example:

        | Open Available Browser | https://google.com |
        | ${res} | Does Page Contain Image | Google |
        """
        return self._run_should_keyword_and_return_status(
            self.page_should_contain_image, locator
        )

    @keyword
    def does_page_contain_link(self, locator: Locator) -> bool:
        """Does page contain expected link

        ``locator`` element locator

        Example:

        | ${res} | Does Page Contain Link | id:submit |
        """
        return self._run_should_keyword_and_return_status(
            self.page_should_contain_link, locator
        )

    @keyword
    def does_page_contain_list(self, locator: Locator) -> bool:
        """Does page contain expected list

        ``locator`` element locator

        Example:

        | ${res} | Does Page Contain List | class:selections |
        """
        return self._run_should_keyword_and_return_status(
            self.page_should_contain_list, locator
        )

    @keyword
    def does_page_contain_radio_button(self, locator: Locator) -> bool:
        """Does page contain expected radio button

        ``locator`` element locator

        Example:

        | ${res} | Does Page Contain Radio Button | male |
        """
        return self._run_should_keyword_and_return_status(
            self.page_should_contain_radio_button, locator
        )

    @keyword
    def does_page_contain_textfield(self, locator: Locator) -> bool:
        """Does page contain expected textfield

        ``locator`` element locator

        Example:

        | ${res} | Does Page Contain Textfield | id:address |
        """
        return self._run_should_keyword_and_return_status(
            self.page_should_contain_textfield, locator
        )

    @keyword
    def is_radio_button_set_to(self, group_name: str, value: str) -> bool:
        """Is radio button group set to expected value

        ``group_name`` radio button group name

        ``value`` expected value

        Example:

        | ${res} | Is Radio Button Set To | group_name=gender | value=female |
        """
        return self._run_should_keyword_and_return_status(
            self.radio_button_should_be_set_to, group_name, value
        )

    @keyword
    def is_radio_button_selected(self, group_name: str) -> bool:
        """Is any radio button selected in the button group

        ``group_name`` radio button group name

        Example:

        | ${res} | Is Radio Button Selected | group_name=gender |
        """
        self.logger.info(
            "Will return if anything is selected on the radio button group"
        )
        return not self._run_should_keyword_and_return_status(
            self.radio_button_should_not_be_selected, group_name
        )

    @keyword
    def does_table_cell_contain(
        self, locator: Locator, row: int, column: int, expected: str
    ) -> bool:
        """Does table cell contain expected text

        ``locator`` element locator for the table

        ``row`` row index starting from 1 (beginning) or -1 (from the end)

        ``column`` column index starting from 1 (beginning) or -1 (from the end)

        ``expected`` expected text in table row

        Example:

        | ${res} | Does Table Cell Contain | //table | 1 | 1 | Company |
        """
        return self._run_should_keyword_and_return_status(
            self.table_cell_should_contain, locator, row, column, expected
        )

    @keyword
    def does_table_column_contain(
        self, locator: Locator, column: int, expected: str
    ) -> bool:
        """Does table column contain expected text

        ``locator`` element locator for the table

        ``column`` column index starting from 1 (beginning) or -1 (from the end)

        ``expected`` expected text in table column

        Example:

        | ${res} | Does Table Column Contain | //table | 1 | Nokia |
        """
        return self._run_should_keyword_and_return_status(
            self.table_column_should_contain, locator, column, expected
        )

    @keyword
    def does_table_row_contain(self, locator: Locator, row: int, expected: str) -> bool:
        """Does table row contain expected text

        ``locator`` element locator for the table

        ``row`` row index starting from 1 (beginning) or -1 (from the end)

        ``expected`` expected text in table row

        Example:

        | ${res} | Does Table Row Contain | //table | 1 | Company |
        """
        return self._run_should_keyword_and_return_status(
            self.table_row_should_contain, locator, row, expected
        )

    @keyword
    def does_table_footer_contain(self, locator: Locator, expected: str) -> bool:
        """Does table footer contain expected text

        ``locator`` element locator for the table

        ``expected`` expected text in table footer

        Example:

        | ${res} | Does Table Footer Contain | //table | Sum |
        """
        return self._run_should_keyword_and_return_status(
            self.table_footer_should_contain, locator, expected
        )

    @keyword
    def does_table_header_contain(self, locator: Locator, expected: str) -> bool:
        """Does table header contain expected text

        ``locator`` element locator for the table

        ``expected`` expected text in table header

        Example:

        | ${res} | Does Table Header Contain | //table | Month |
        """
        return self._run_should_keyword_and_return_status(
            self.table_header_should_contain, locator, expected
        )

    @keyword
    def does_table_contain(self, locator: Locator, expected: str) -> bool:
        """Does table contain expected text

        ``locator`` element locator

        ``expected`` expected text in table

        Example:

        | ${res} | Does Table Contain | //table | February |
        """
        return self._run_should_keyword_and_return_status(
            self.table_should_contain, locator, expected
        )

    @keyword
    def is_textarea_value(self, locator: Locator, expected: str) -> bool:
        """Is textarea matching expected value

        ``locator`` element locator

        ``expected`` expected textarea value

        Example:

        | ${res} | Is Textarea Value | //textarea | Yours sincerely |
        """
        return self._run_should_keyword_and_return_status(
            self.textarea_value_should_be, locator, expected
        )

    @keyword
    def does_textarea_contain(self, locator: Locator, expected: str) -> bool:
        """Does textarea contain expected text

        ``locator`` element locator

        ``expected`` expected text in textarea

        Example:

        | ${res} | Does Textarea Contain | //textarea | sincerely |
        """
        return self._run_should_keyword_and_return_status(
            self.textarea_should_contain, locator, expected
        )

    @keyword
    def does_textfield_contain(self, locator: Locator, expected: str) -> bool:
        """Does textfield contain expected text

        ``locator`` element locator

        ``expected`` expected text in textfield

        Example:

        | ${res} | Does Textfield Contain | id:lname | Last |
        """
        return self._run_should_keyword_and_return_status(
            self.textfield_should_contain, locator, expected
        )

    @keyword
    def is_textfield_value(self, locator: Locator, expected: str) -> bool:
        """Is textfield value expected

        ``locator`` element locator

        ``expected`` expected textfield value

        Example:

        | ${res} | Is Textfield Value | id:lname | Lastname |
        """
        return self._run_should_keyword_and_return_status(
            self.textfield_value_should_be, locator, expected
        )

    @keyword
    def is_title(self, title: str) -> bool:
        """Is page title expected

        ``title`` expected title value

        Example:

        | ${res} | Is Title | Webpage title text |
        """
        return self._run_should_keyword_and_return_status(self.title_should_be, title)

    def _run_should_keyword_and_return_status(self, runnable_keyword, *args, **kwargs):
        missing_ok = kwargs.pop("missing_ok", False)
        catches = (AssertionError, ElementNotFound) if missing_ok else (AssertionError)

        try:
            runnable_keyword(*args, **kwargs)
            return True
        except catches as e:
            try:
                BuiltIn().log(
                    "Ran with keyword <b>%s</b> which returned error: <i>%s</i>"
                    % (runnable_keyword.__func__.__name__.replace("_", " ").title(), e),
                    html=True,
                )
            except RobotNotRunningError:
                pass
            return False

    @keyword
    def get_element_status(self, locator: Locator) -> dict:
        """Return dictionary containing element status of:

            - visible
            - enabled
            - disabled
            - focused

        ``locator`` element locator

        Example:

        | &{res}  | Get Element Status | class:special |
        | Log     | ${res.visible} |
        | Log     | ${res.enabled} |
        | Log     | ${res.disabled} |
        | Log     | ${res.focused} |
        """
        status_object = {}
        status_object["visible"] = self.is_element_visible(locator)
        status_object["enabled"] = self.is_element_enabled(locator)
        status_object["disabled"] = self.is_element_disabled(locator)
        status_object["focused"] = self.is_element_focused(locator)
        notebook.notebook_json(status_object)
        return status_object

    @keyword
    def get_testability_status(self) -> bool:
        """Get SeleniumTestability plugin status"""
        return self.using_testability

    @keyword
    def open_user_browser(self, url: str, tab=True) -> None:
        """Opens an URL with te user's default browser.

        The browser opened with this keyword is not accessible
        with Selenium. To interact with the opened browser it is
        possible to use ``RPA.Desktop`` or ``RPA.Windows`` library keywords.

        The keyword `Attach Chrome Browser` can be used to
        access an already open browser with Selenium keywords.

        Read more: https://robocorp.com/docs/development-guide/browser/how-to-attach-to-running-chrome-browser

        ``url`` URL to open
        ``tab`` defines is url is opened in a tab (defaults to ``True``) or
                in new window (if set to ``False``)

        Example:

        | Open User Browser  | https://www.google.com?q=rpa |
        | Open User Browser  | https://www.google.com?q=rpa | tab=${False} |
        """  # noqa: E501
        browser_method = webbrowser.open_new_tab if tab else webbrowser.open_new
        browser_method(url)

    @keyword
    def get_browser_capabilities(self) -> dict:
        """Get dictionary of browser properties

        Example:

        | ${caps}= | Get Browser Capabilities |
        """
        capabilities = self.driver.capabilities
        return dict(capabilities)

    @keyword
    def set_download_directory(
        self, directory: Optional[str] = None, download_pdf: bool = True
    ) -> None:
        """Set a custom browser download directory.

        This has to be called before opening the browser and it works with the
        following keywords:

        - ``Open Available Browser``
        - ``Open Chrome Browser``
        - ``Open Headless Chrome Browser``

        Supported browsers: Chrome, Edge, Firefox.

        If the downloading doesn't work (file is not found on disk), try using the
        browser in non-headless (headful) mode when opening it. (``headless=${False}``)

        Parameter ``directory`` sets a path for downloads, defaults to ``None``, which
        means that this setting is removed and the default location will be used.
        Parameter ``download_pdf`` will download a PDF file instead of previewing it
        within browser's internal viewer when this is set to ``True``. (enabled by
        default)

        Example:

        | `Set Download Directory` | ${OUTPUT_DIR}           |
        | Open Available Browser   | https://cdn.robocorp.com/legal/Robocorp-EULA-v1.0.pdf |
        | @{files} =               | List Files In Directory | ${OUTPUT_DIR}               |
        | Log List                 |  ${files}               |
        """  # noqa: E501
        if directory is None:
            self.logger.info(
                "Download directory set back to browser's default setting!"
            )
            self.download_preferences.clear()
            return

        download_directory = Path(directory).expanduser().resolve()
        download_directory.mkdir(parents=True, exist_ok=True)
        download_directory = str(download_directory)
        self.logger.info("Download directory set to: %s", download_directory)
        chromium_prefs = {
            "download.default_directory": download_directory,
            "plugins.always_open_pdf_externally": download_pdf,
            "download.directory_upgrade": True,
            "download.prompt_for_download": False,
        }
        firefox_prefs = {
            "browser.download.folderList": 2,
            "browser.download.manager.showWhenStarting": False,
            "browser.download.improvements_to_download_panel": True,
            "browser.download.useDownloadDir": True,
            "browser.helperApps.alwaysAsk.force": False,
            "pdfjs.disabled": True,
            "browser.download.dir": download_directory,
            # MIME types: https://www.freeformatter.com/mime-types-list.html
            "browser.helperApps.neverAsk.saveToDisk": "application/octet-stream",
        }
        if download_pdf:
            # Disable the viewer when downloading is preferred instead of viewing.
            firefox_prefs["browser.download.viewableInternally.enabledTypes"] = ""
            firefox_prefs[
                "browser.helperApps.neverAsk.saveToDisk"
            ] += " application/pdf"
        self.download_preferences = {
            "firefox": firefox_prefs,
        }
        for browser_lower in self.CHROMIUM_BROWSERS:
            self.download_preferences[browser_lower] = chromium_prefs

    @keyword
    def highlight_elements(
        self,
        locator: Locator,
        width: str = "2px",
        style: str = "dotted",
        color: str = "blue",
    ):
        """
        Highlight all matching elements by locator.

        Highlighting is done by adding a colored outline
        around the elements with CSS styling.

        ``locator``  element locator
        ``width``    highlight outline width
        ``style``    highlight outline style
        ``color``    highlight outline color

        Example:

        | Highlight Elements | xpath://h2 |
        """
        elements = self.find_elements(locator)
        attribute_name = "rpaframework-highlight"

        def inject_style():
            css = (
                "\n"
                f"[{attribute_name}] {{\n"
                f"  outline: {width} {style} {color};\n"
                "}\n"
            )
            script = (
                "var node = document.createElement('style');\n"
                "node.setAttribute('data-name', 'rpaframework');\n"
                f"node.innerHTML=`{css}`\n"
                "document.head.appendChild(node);"
            )
            self.driver.execute_script(script)

        def add_highlight_attribute_to_elements():
            script = "".join(
                f'arguments[{idx}].setAttribute("{attribute_name}", "");'
                for idx in range(len(elements))
            )
            self.driver.execute_script(script, *elements)

        inject_style()
        add_highlight_attribute_to_elements()

    @keyword
    def clear_all_highlights(self):
        """Remove all highlighting made by ``Highlight Elements``."""
        attribute_name = "rpaframework-highlight"

        elements = self.driver.find_elements(By.CSS_SELECTOR, f"[{attribute_name}]")
        script = "".join(
            f'arguments[{idx}].removeAttribute("{attribute_name}");'
            for idx in range(len(elements))
        )
        self.driver.execute_script(script, *elements)

    @property
    def is_chromium(self) -> bool:
        return self.driver.name.lower() in self.CHROMIUM_BROWSERS

    @keyword
    def print_to_pdf(
        self, output_path: Optional[str] = None, params: Optional[dict] = None
    ) -> str:
        """Print the current page to a PDF document using Chrome's DevTools.

        Attention: With some older browsers, this may work in *headless* mode only!
        For a list of supported parameters see:
        https://chromedevtools.github.io/devtools-protocol/tot/Page/#method-printToPDF
        Returns the output PDF file path.

        Parameter ``output_path`` specifies the file path for the generated PDF
        document. By default, it is saved to the output folder with the default name
        of `out.pdf`.
        Parameter ``params`` specify parameters for the browser printing method. By
        default, it uses the following values:
        ```
        {
            "landscape": False,
            "displayHeaderFooter": False,
            "printBackground": True,
            "preferCSSPageSize": True,
        }
        ```
        """
        if not self.is_chromium:
            raise NotImplementedError(
                f"PDF printing works only with Chromium-based browsers,"
                f" got: {self.driver.name}"
            )

        default_params = {
            "landscape": False,
            "displayHeaderFooter": False,
            "printBackground": True,
            "preferCSSPageSize": True,
        }
        params = params or default_params
        result = self._send_command_and_get_result("Page.printToPDF", params)
        if isinstance(result, str) and "Printing is not available" in result:
            raise TypeError("PDF printing works in headless mode only")

        output_path = output_path or str(get_output_dir() / "out.pdf")
        pdf_data = base64.b64decode(result["data"])
        with open(output_path, "wb") as stream:
            stream.write(pdf_data)

        return output_path

    @keyword
    def execute_cdp(self, command, parameters):
        """
        Executes Chromium DevTools Protocol commands

        Works only with Chromium-based browsers!

        For more information, available commands and parameters, see:
        https://chromedevtools.github.io/devtools-protocol/

        ``command`` command to execute as string

        ``parameters`` parameters for command as a dictionary

        Example:

        | Open Chrome Browser | about:blank | headless=${True} |
        | &{params} | Create Dictionary | userAgent=Chrome/83.0.4103.53 |
        | Execute CDP | Network.setUserAgentOverride | ${params} |
        | Go To | https://robocorp.com |
        """
        if not self.is_chromium:
            raise NotImplementedError(
                "Executing DevTools Protocol commands"
                f" works only with Chromium-based browsers, got: {self.driver.name}",
            )
        return self._send_command_and_get_result(command, parameters)

    def _send_command_and_get_result(self, cmd, params):
        resource = (
            f"session/{self.driver.session_id}/chromium/send_command_and_get_result"
        )
        # pylint: disable=protected-access
        url = f"{self.driver.command_executor._url}/{resource}"
        body = json.dumps({"cmd": cmd, "params": params})
        response = self.driver.command_executor._request("POST", url, body)

        return response.get("value")

    @keyword
    def set_element_attribute(
        self, locator: Locator, attribute: str, value: str
    ) -> None:
        """Sets a ``value`` for the ``attribute`` in the element ``locator``.

        See the `Locating elements` section for details about the locator
        syntax.

        Example:

        | Set Element Attribute | css:h1 | class | active |
        """
        element = self.find_element(locator)
        # NOTE(cmin764): The `WebElement` object doesn't support the `set_attribute`
        #  method nor the execution of the "setElementAttribute" command in order to
        #  canonically set a `value` to the passed `attribute`.
        #  Therefore we execute a script instead:
        script = "arguments[0].setAttribute(arguments[1], arguments[2]);"
        element.parent.execute_script(script, element, attribute, value)

    @keyword
    def click_element_when_clickable(
        self, locator: Locator, timeout: TimeoutType = None
    ) -> None:
        """Waits for and clicks an element until is fully ready to be clicked.

        If a normal click doesn't work, then JavaScript-oriented workarounds are tried
        as a fallback mechanism.

        Parameter ``locator`` targets the element to be clicked.
        Parameter ``timeout`` optionally configures a custom duration to wait for the
        element to become clickable, until it gives up.

        Example:

        | Click Element When Clickable | example |
        """
        element = self.find_element(locator)
        timeout: float = self.browser_management.get_timeout(timeout)
        wait = WebDriverWait(self.driver, timeout)
        clickable_element = wait.until(
            expected_conditions.element_to_be_clickable(element)
        )
        try:
            clickable_element.click()
        except ElementClickInterceptedException as exc:
            self.logger.warning(
                "Couldn't click element %r with Selenium due to: %s"
                " (trying with JavaScript)",
                locator,
                exc,
            )
            self.driver.execute_script("arguments[0].click();", clickable_element)

    @keyword(name="Get WebElement")
    def get_webelement(
        self, locator: Locator, parent: Optional[Element] = None, shadow: bool = False
    ) -> Element:
        """Returns the first ``Element`` matching the given ``locator``.

        With the ``parent`` parameter you can optionally specify a parent to start the
        search from. Set ``shadow`` to ``True`` if you're targeting and expecting a
        shadow root in return. Read more on the shadow root:
        https://developer.mozilla.org/en-US/docs/Web/API/ShadowRoot

        See the `Locating elements` section for details about the locator
        syntax.
        """
        element = self.find_element(locator, parent=parent)
        return element.shadow_root if shadow else element



import pytest


class TestOutlook:
    """Gathers `RPA.Outlook.Application` tests."""

    @pytest.fixture
    def Application(self):
        from RPA.Outlook.Application import Application

        return Application

    @pytest.fixture
    def app(self, Application):
        return Application()

    @pytest.fixture
    def custom_app(self, Application):
        class CustomApplication(Application):
            """Some test docstring."""

        return CustomApplication()

    @pytest.fixture
    def custom_app_no_docs(self, Application):
        class CustomApplication(Application):
            pass

        return CustomApplication()

    def test_import(self, app):
        assert app.__doc__.startswith("`Outlook.Application`")

    def test_custom_import(self, custom_app):
        assert custom_app.__doc__.startswith("Some test docstring.")

    def test_custom_import_no_docs(self, custom_app_no_docs):
        assert custom_app_no_docs.__doc__.startswith("`Outlook.Application`")
